# Built-in
import logging
import re
from code import InteractiveConsole
import sys
import os
import time

# External
from Qt import QtWidgets, QtGui, QtCore

# Internal
import nxt_editor
from nxt_editor import user_dir
from nxt_editor.dockwidgets.dock_widget_base import DockWidgetBase
from nxt import nxt_log
from nxt_editor import LoggingSignaler, colors

logger = logging.getLogger(nxt_editor.LOGGER_NAME)


class VisualLogHandler(logging.Handler):
    def __init__(self, output_log):
        logging.Handler.__init__(self)
        self.setFormatter(nxt_log.nxt_formatter)
        self.signaller = LoggingSignaler()
        self.signaller.signal.connect(self.update)
        self.target_output_log = output_log

    @staticmethod
    def format_links(message, links):
        """Given a message string and a list of links, replace every occurance
        of each given link in the message with a link.

        :param message: string containing unformatted links
        :type message: str
        :param links: list of strings to format into links
        :type links: list
        :return: message with links replacing original text
        :rtype: str
        """
        # Determine max length and sort links by length
        links_by_length = {}
        max_length = 0
        for link_str in links:
            length = len(link_str)
            links_by_length.setdefault(length, [])
            links_by_length[length].append(link_str)
            if length > max_length:
                max_length = length
        # search_message is always length of original string. When searching,
        # found link_str are replaced with % to prevent links within links.
        search_message = str(message)
        out_mesage = ''
        # Gather matches list that is ordered from left to right
        matches = []
        for i in reversed(range(max_length + 1)):
            for link_str in links_by_length.get(i, []):
                length = len(link_str)
                new_search = search_message
                for new_match in re.finditer(link_str, search_message):
                    start = new_match.start()
                    end = new_match.end()
                    filler = length*'%'
                    new_search = new_search[:start] + filler + new_search[end:]
                    # Order matches from left to right based on start index
                    if len(matches) == 0:
                        matches.append(new_match)
                    else:
                        for m in matches:
                            if start < m.start():
                                matches.insert(0, new_match)
                                break
                        else:
                            matches.append(new_match)
                search_message = new_search
        # assemble formatted message using indices of matches.
        prev_idx = 0
        for match in matches:
            out_mesage += message[prev_idx:match.start()]
            out_mesage += nxt_log.make_link(match.re.pattern)
            prev_idx = match.end()
        out_mesage += message[prev_idx:]

        return out_mesage

    def format(self, record):
        if isinstance(record.msg, tuple):
            text = record.msg[0]
            original_msg = record.msg
            multi = True
        else:
            text = record.msg
            original_msg = text
            multi = False
        try:
            links = record.links
        except AttributeError:
            links = []
        if links:
            text = self.format_links(text, links)
        msg = '<font face="Roboto Mono" color="white">{}</font>'.format(text)
        if multi:
            replacement = (msg, record.msg[1])
        else:
            replacement = msg
        record.msg = replacement
        formatted = super(VisualLogHandler, self).format(record)
        record.msg = original_msg
        return formatted

    def emit(self, record):
        self.signaller.signal.emit(record)

    def update(self, record):
        text = self.format(record)
        text += '\n'
        self.target_output_log.write_rich_output(text, level=record.levelno)


class WriteDuper(QtCore.QObject):
    message_written = QtCore.Signal(str, float)

    def __init__(self, orig_obj):
        """Given an object to wrap, and a method to use as write. Produces an
        object that can be used as a proxy for the original to duplicate
        write calls both to given write function as well as to original object.

        Designed for wrapping sys.stdout and sys.stderr to duplicate output.

        :param orig_obj: Object to produce wrapper for
        :type orig_obj: object
        """
        super(WriteDuper, self).__init__()
        self._orig_obj = orig_obj

    def __getattr__(self, attr):
        if attr in self.__dict__:
            return getattr(self, attr)
        return getattr(self._orig_obj, attr)

    def write(self, val):
        self.message_written.emit(val, time.time())
        self._orig_obj.write(val)


class FileTailingThread(QtCore.QThread):
    """A QThread to continuously monitor a file path and signal when text
    is added. On initial start, if the watched file has existing text, the
    the first signal will contain all the initial text.
    """
    new_text = QtCore.Signal(str)

    def __init__(self, path):
        """Watch given path

        :param path: [description]
        :type path: [type]
        """
        super(FileTailingThread, self).__init__()
        self.watch_path = path
        self.last_mtime = 0
        self.last_read_pos = 0

    def run(self):
        while not self.isInterruptionRequested():
            # Consider using a QTimer instead of a thread... -LB
            time.sleep(.2)
            try:
                current_mtime = os.path.getmtime(self.watch_path)
            except OSError:
                # We consume this error because we're watching a file path
                # and the file may be deleted and re-created at any moment.
                continue
            if current_mtime == self.last_mtime:
                continue
            self.last_mtime = current_mtime
            with open(self.watch_path, 'r') as fp:
                fp.seek(self.last_read_pos)
                new_text = fp.read()
            self.last_read_pos += len(new_text)
            self.new_text.emit(new_text)


class OutputLog(DockWidgetBase):
    write_raw = QtCore.Signal(str, float)

    def __init__(self, graph_model=None, parent=None):
        super(OutputLog, self).__init__('Output Log', graph_model=graph_model,
                                        minimum_height=100, parent=parent)
        self.main_frame = QtWidgets.QFrame(self)
        self.main_frame.setStyleSheet('background-color: #3E3E3E; '
                                      'border-radius: 0px;')
        self.main_layout = QtWidgets.QVBoxLayout()
        self.main_layout.setContentsMargins(0, 0, 0, 0)
        self.main_layout.setSpacing(0)
        self.setWidget(self.main_frame)
        self.main_frame.setLayout(self.main_layout)

        # Rich Output
        self.rich_output_textedit = OutputTextBrowser(self)
        self.log_filter_button = LogFilterButton()
        self.clear_rich_button = QtWidgets.QPushButton('Clear Log')
        self.clear_rich_button.pressed.connect(self.rich_output_textedit.clear)

        self.buttons_layout = QtWidgets.QHBoxLayout()
        self.buttons_layout.addWidget(self.log_filter_button)
        self.buttons_layout.addStretch(stretch=1)
        self.buttons_layout.addWidget(self.clear_rich_button)

        self.rich_output_layout = QtWidgets.QVBoxLayout()
        self.rich_output_layout.setContentsMargins(0, 0, 0, 0)
        self.rich_output_layout.setSpacing(0)
        self.rich_output_layout.addWidget(self.rich_output_textedit)
        self.rich_output_layout.addSpacing(3)
        self.rich_output_layout.addLayout(self.buttons_layout)
        self.rich_output_page = QtWidgets.QWidget()
        self.rich_output_page.setLayout(self.rich_output_layout)
        # Raw Output
        self.raw_output_textedit = OutputTextEdit(self)
        console_locals = {'main_window': parent}
        self.python_edit = PythonConsoleLineEdit(locals=console_locals)

        self.python_layout = QtWidgets.QHBoxLayout()
        self.python_layout.addWidget(QtWidgets.QLabel("Python: "))
        self.python_layout.addWidget(self.python_edit, stretch=1)
        self.clear_raw_button = QtWidgets.QPushButton('Clear Log')
        self.clear_raw_button.pressed.connect(self.raw_output_textedit.clear)
        self.python_layout.addWidget(self.clear_raw_button)

        self.raw_output_layout = QtWidgets.QVBoxLayout()
        self.raw_output_layout.setContentsMargins(0, 0, 0, 0)
        self.raw_output_layout.setSpacing(0)
        self.raw_output_layout.addWidget(self.raw_output_textedit)
        self.raw_output_layout.addSpacing(3)
        self.raw_output_layout.addLayout(self.python_layout)
        self.raw_output_page = QtWidgets.QWidget()
        self.raw_output_page.setLayout(self.raw_output_layout)
        self.write_raw.connect(self.write_raw_output)
        # Tabs
        self.tabs_widget = QtWidgets.QTabWidget()
        self.tabs_widget.addTab(self.rich_output_page, 'Rich Output')
        self.tabs_widget.addTab(self.raw_output_page, 'Raw Output')

        self.main_layout.addWidget(self.tabs_widget)

        self.log_watcher = None
        self.wrapped_stdout = None
        self.wrapped_stderr = None
        self.removed_stderr_handler = None
        self.replaced_stderr_handler = None
        self.wrap_std_streams_for_raw_output()

        # Catch-up visual log to file log
        with open(self.parent().nxt.log_file, 'r') as fp:
            catch_up = fp.read()
        catch_up = catch_up.rstrip()  # removing hanging newlines.
        self.rich_output_textedit.append(catch_up)

        # Install visual handler to output nxt logging to rich log
        self.visual_handler = VisualLogHandler(self)
        self.visual_handler.addFilter(self.log_filter_button)
        nxt_logger = logging.getLogger('nxt')
        nxt_logger.addHandler(self.visual_handler)

    def wrap_std_streams_for_raw_output(self):
        if self.log_watcher:
            self.log_watcher.requestInterruption()
        if self.wrapped_stderr or self.wrapped_stdout:
            logger.debug("std streams already wrapped, not re-wrapping")
            return
        # Make our "Raw Output" log stand in for stdout and stderr
        self.wrapped_stdout = WriteDuper(sys.stdout)
        self.wrapped_stdout.message_written.connect(self.write_raw_output)
        sys.stdout = self.wrapped_stdout
        self.wrapped_stderr = WriteDuper(sys.stderr)
        self.wrapped_stderr.message_written.connect(self.write_raw_output)
        sys.stderr = self.wrapped_stderr

        # Nxt log maintainence
        # nxt's logging goes out stderr and has started before output log,
        # replace existing nxt stdout/err handlers.
        nxt_logger = logging.getLogger('nxt')
        replace_handler = None
        for handler in nxt_logger.handlers:
            if not isinstance(handler, logging.StreamHandler):
                continue
            if handler.stream == sys.__stderr__:
                replace_handler = handler
                break
        if replace_handler:
            self.removed_stderr_handler = replace_handler
            nxt_logger.removeHandler(handler)

            stderr_handler = logging.StreamHandler(self.wrapped_stderr)
            stderr_handler.setLevel(replace_handler.level)
            self.replaced_stderr_handler = stderr_handler
            stderr_handler.setFormatter(nxt_log.nxt_formatter)
            nxt_logger.addHandler(stderr_handler)

    def unwrap_std_streams(self):
        if not (self.wrapped_stdout and self.wrapped_stderr):
            # Not wrapped
            return
        # Wrapping
        sys.stdout = self.wrapped_stdout._orig_obj
        sys.stderr = self.wrapped_stderr._orig_obj
        # Nxt log maintainence
        if self.replaced_stderr_handler:
            nxt_logger = logging.getLogger('nxt')
            nxt_logger.removeHandler(self.replaced_stderr_handler)
            self.replaced_stderr_handler = None
        if self.removed_stderr_handler:
            nxt_logger.addHandler(self.removed_stderr_handler)
            self.removed_stderr_handler = None

    def write_raw_output(self, val, msg_time=0.):
        """Write text to the raw output textedit. May write to the rich out
        if there is a stage model and its current runtime layer is actively
        running.
        The use of the word write here is specific, no newlines are
        added implicitly, must be added
        via input.
        :param msg_time: Optional time when the message was processed,
        if no is given the val will never be logged to the rich log.
        :param val: Text to add to textedit
        :type val: str
        """
        self._write_raw_output(val)
        if not self.stage_model:
            return
        curr_rt_layer = self.stage_model.current_rt_layer
        if not (curr_rt_layer and
                curr_rt_layer.cache_layer.was_during_node_exec(msg_time)):
            return
        if self.log_filter_button.is_level_enabled(nxt_log.NODEOUT):
            # Intentionally breaking html syntax here. Printing type(object)
            # was being interpreted as html. <class 'type'>
            self.write_rich_output(val.replace('<', '&lt;'))

    def _write_raw_output(self, val):
        """Write text to the raw output textedit. Dose NOT write to the rich
        output. The use of the word write here is specific, no newlines are
        added implicitly, must be added via input.
        :param val: Text to add to textedit
        :type val: str
        """
        _max = self.raw_output_textedit.verticalScrollBar().maximum()
        cur = self.raw_output_textedit.verticalScrollBar().value()
        self.raw_output_textedit.moveCursor(QtGui.QTextCursor.End)
        self.raw_output_textedit.insertPlainText(val)
        if _max == cur:
            new_max = self.raw_output_textedit.verticalScrollBar().maximum()
            self.raw_output_textedit.verticalScrollBar().setValue(new_max)
        else:
            self.raw_output_textedit.verticalScrollBar().setValue(cur)

    def write_rich_output(self, val, level=None):
        """Neighbor to write_raw_output. Use of write as naming is specific,
        no newlines created from thin air, must come form input.

        :param val: text to add
        :type val: str
        :param level: logging level to color text as, defaults to None
        :type level: int, optional
        """
        _max = self.rich_output_textedit.verticalScrollBar().maximum()
        cur = self.rich_output_textedit.verticalScrollBar().value()
        self.rich_output_textedit.moveCursor(QtGui.QTextCursor.End)
        if val == '\n':
            self.rich_output_textedit.append('')
        else:
            final_newline = False
            if val.endswith('\n'):
                final_newline = True
                text = val[:-1]
            else:
                text = val
            color = colors.LOGGING_COLORS.get(level, 'white')
            html = '<font face="Roboto Mono" color="{}">'.format(color)
            style = ("<style type='text/css'> "
                     "pre {margin: 0; font-family: 'Roboto Mono';} "
                     "</style>")
            text = style + ("<pre>{}</pre>".format(text))
            html += text + '</font>'
            self.rich_output_textedit.insertHtml(html)
            if final_newline:
                self.rich_output_textedit.append('')
        if _max == cur:
            new_max = self.rich_output_textedit.verticalScrollBar().maximum()
            self.rich_output_textedit.verticalScrollBar().setValue(new_max)
        else:
            self.rich_output_textedit.verticalScrollBar().setValue(cur)

    def tail_file_for_raw_output(self, file_path):
        """Treats the given file as what should be displayed in "raw output"
        Spawns a thread that continually watches given file for modifictions
        and appends new lines to the end of raw output. If this has not been
        called before, and nxt log has wrapped stdout, stdout will be restored
        to default and given file will be appended below. If this has been
        called before, successive calls will replace existing file, and
        continue to append below existing log.

        :param file_path: file to treat as raw output
        :type file_path: str
        """
        self.unwrap_std_streams()
        if self.log_watcher:
            self.log_watcher.requestInterruption()
        self.log_watcher = FileTailingThread(file_path)
        self.log_watcher.new_text.connect(self.write_raw_output)
        self.log_watcher.start()

    def link_clicked(self, link):
        view = self.parent().get_current_view()
        if not view:
            return
        model = view.model
        link = link.toString()
        model.select_and_frame(link)

    def closeEvent(self, event):
        if self.log_watcher:
            self.log_watcher.requestInterruption()
        super(OutputLog, self).closeEvent(event)


class OutputTextEdit(QtWidgets.QTextEdit):
    def __init__(self, parent):
        super(OutputTextEdit, self).__init__(parent=parent)
        self._parent = parent
        self.setStyleSheet(self.parent().parent().styleSheet())
        self.setReadOnly(True)
        self.setFont(QtGui.QFont('Roboto Mono', 10))

    def contextMenuEvent(self, event):
        menu = self.createStandardContextMenu()
        menu.addAction('Clear', self.clear)
        menu.addAction(self._parent.parent().app_actions.clear_logs_action)
        menu.exec_(event.globalPos())


class OutputTextBrowser(QtWidgets.QTextBrowser):
    def __init__(self, parent):
        super(OutputTextBrowser, self).__init__(parent=parent)
        self._parent = parent
        self.anchorClicked.connect(self.parent().link_clicked)
        self.setStyleSheet(self.parent().parent().styleSheet())
        self.setOpenLinks(False)
        self.setFont(QtGui.QFont('Roboto Mono', 10))

    def contextMenuEvent(self, event):
        menu = self.createStandardContextMenu()
        menu.addAction('Clear', self.clear)
        menu.addAction(self._parent.parent().app_actions.clear_logs_action)
        menu.exec_(event.globalPos())


class LogFilterButton(QtWidgets.QPushButton):
    VISUAL_LOG_LEVELS = [
        logging.DEBUG,
        logging.INFO,
        logging.WARNING,
        logging.ERROR,
        logging.CRITICAL,
        nxt_log.NODEOUT,
        nxt_log.COMPINFO,
        nxt_log.EXECINFO,
        nxt_log.GRAPHERROR,
        nxt_log.SOCKET
    ]
    PREF_KEY = user_dir.USER_PREF.LOG_FILTERS

    def __init__(self, *args, **kwargs):
        super(LogFilterButton, self).__init__('Filter Log', *args, **kwargs)
        self.menu = QtWidgets.QMenu()
        self.lvl_actions = {}
        for lvl in self.VISUAL_LOG_LEVELS:
            lvl_name = logging.getLevelName(lvl)
            new_action = self.menu.addAction(lvl_name)
            new_action.setCheckable(True)
            new_action.setChecked(True)
            self.lvl_actions[lvl] = new_action
        self.all_action = self.menu.addAction('All')
        self.all_action.setCheckable(True)
        self.all_action.setChecked(True)
        self.setMenu(self.menu)
        self.load_filters_from_pref()
        self.menu.triggered.connect(self.action_triggered)

    def filter(self, record):
        """"filter" for use in logging handlers.
        Name is critical, as logging calls `inst.filter` on given instance.
        """
        return self.is_level_enabled(record.levelno)

    def is_level_enabled(self, lvl):
        action = self.lvl_actions.get(lvl, None)
        if not action:
            # If we don't know we don't care.
            return True
        return action.isChecked()

    def load_filters_from_pref(self):
        pref_filters = user_dir.user_prefs.get(self.PREF_KEY)
        if not pref_filters:
            return
        for lvl, checked in pref_filters.items():
            lvl = int(lvl)
            action = self.lvl_actions.get(lvl)
            if not action:
                continue
            action.setChecked(checked)

    def action_triggered(self, action):
        if action == self.all_action:
            for other_action in self.lvl_actions.values():
                # don't want to recurse this method.
                other_action.setChecked(action.isChecked())
        self.save_filters_to_pref()

    def save_filters_to_pref(self):
        """Saves filter settings to a user preference.
        """
        output = {}
        for lvl, action in self.lvl_actions.items():
            output[lvl] = action.isChecked()
        user_dir.user_prefs[self.PREF_KEY] = output


class PythonConsoleLineEdit(QtWidgets.QLineEdit):
    def __init__(self, locals={}):
        super(PythonConsoleLineEdit, self).__init__('')
        self.console = InteractiveConsole(locals, '<nxt console>')
        self.returnPressed.connect(self.on_return)

    def on_return(self):
        need_more = self.console.push(self.text())
        if not need_more:
            self.clear()


class QtLogStreamHandler(nxt_log.LogRecordStreamHandler):
    """Handles logs by emitting the log record to a QtCore.Signal"""
    @classmethod
    def get_handler(cls, signal):
        """Get the QtLogStreamHandler class with the provided signal
        assigned to it. Meant to only be called once per UI session.
        :param signal: QtCore.Signal that takes a logging.LogRecord
        :return: QtLogStreamHandler
        """
        cls.new_log = signal
        return cls

    def handle_log_record(self, record):
        self.new_log.emit(record)



# External
from Qt import QtWidgets
from Qt import QtGui
from Qt import QtCore


class PixmapButton(QtWidgets.QAbstractButton):
    """https://stackoverflow.com/questions/2711033/how-code-a-image-button-in-pyqt"""
    def __init__(self, pixmap, pixmap_hover, pixmap_pressed, pixmap_checked=None,
                 pixmap_checked_hover=None, pixmap_checked_pressed=None, size=32, checkable=False,
                 parent=None, resize_signal=None):
        super(PixmapButton, self).__init__(parent=parent)
        self.pixmap = pixmap
        self.pixmap_hover = pixmap_hover
        self.pixmap_pressed = pixmap_pressed
        self.pixmap_checked = pixmap_checked
        self.pixmap_checked_hover = pixmap_checked_hover
        self.pixmap_checked_pressed = pixmap_checked_pressed
        self.size = size
        self.default_size = size

        if checkable:
            self.setCheckable(checkable)

        self.pressed.connect(self.update)
        self.released.connect(self.update)

        self.action = None
        if resize_signal:
            resize_signal.connect(self.font_size_changed)

    def font_size_changed(self, delta):
        if delta == 0:
            self.size = self.default_size
        else:
            self.size = self.size + delta
        self.setIconSize(QtCore.QSize(self.size, self.size))
        self.update()

    def set_action(self, action):
        self.action = action

        # get properties
        self.setToolTip(self.action.toolTip())
        self.setWhatsThis(self.action.whatsThis())

        # connect signals
        action.triggered.connect(self.update_state)
        action.toggled.connect(self.update_state)
        if action.isCheckable():
            self.toggled.connect(action.toggle)
        else:
            self.clicked.connect(action.trigger)

    def update_state(self):
        if self.action:
            self.blockSignals(True)
            self.setChecked(self.action.isChecked())
            self.blockSignals(False)

    def paintEvent(self, event):
        if not isinstance(event, QtGui.QPaintEvent):
            return
        if self.isChecked():
            pix = self.pixmap_checked_hover if self.underMouse() else self.pixmap_checked
            if self.isDown():
                pix = self.pixmap_checked_pressed
        else:
            pix = self.pixmap_hover if self.underMouse() else self.pixmap
            if self.isDown():
                pix = self.pixmap_pressed

        painter = QtGui.QPainter(self)
        painter.drawPixmap(event.rect(), pix)
        del painter

    def enterEvent(self, event):
        self.update()

    def leaveEvent(self, event):
        self.update()

    def sizeHint(self):
        return QtCore.QSize(self.size, self.size)


"""Depends on the following environment variables being populated.
NXT_ENV_PATH: mapped to site-packages directory containing nxt's dependencies.
NXT_PATH: mapped to a directory containing the nxt package.
"""
# Built-in
import sys
import webbrowser
import logging
import time
import os

# External
# maya
from maya import cmds
from maya import mel
import maya.api.OpenMaya as om
from Qt import QtCore

# Internal
import nxt_editor.main_window
import nxt.remote.nxt_socket
from nxt import nxt_log
from nxt_editor.constants import NXT_WEBSITE
from nxt.constants import NXT_DCC_ENV_VAR

logger = logging.getLogger('nxt')
CREATED_UI = []
global __NXT_INSTANCE__
__NXT_INSTANCE__ = None


class MAYA_PLUGIN_VERSION(object):
    # TODO: Where/if to track these
    # with open(version_file, 'r') as f:
    #     version_data = json.load(f)
    # plugin_v_data = version_data['MAYA_PLUGIN']
    plugin_v_data = {'MAJOR': 0,
                     'MINOR': 1,
                     'PATCH': 0}
    MAJOR = plugin_v_data['MAJOR']
    MINOR = plugin_v_data['MINOR']
    PATCH = plugin_v_data['PATCH']
    VERSION_TUPLE = (MAJOR, MINOR, PATCH)
    VERSION_STR = '.'.join(str(v) for v in VERSION_TUPLE)
    VERSION = VERSION_STR


def about_menu(*args):
    webbrowser.open_new(NXT_WEBSITE)


def auto_reload(*args):
    safe = True
    global __NXT_INSTANCE__
    if __NXT_INSTANCE__:
        safe = __NXT_INSTANCE__.close()
    if safe:
        cmds.nxt_ui('reload')
    else:
        cmds.warning('Aborted reload!')


def enable_cmd_port(enable):
    logger.warning('This is a placeholder!')
    # port = nxt.remote.nxt_socket.CMD_PORT
    # host = nxt.remote.nxt_socket.HOST
    # address = '{}:{}'.format(host, port)
    # if not cmds.commandPort(address, query=True) and enable:
    #     cmds.warning('Opening nxt cmd port...')
    #     cmds.commandPort(name=address, prefix="python",
    #                      sourceType="mel", bs=2048)
    # elif cmds.commandPort(address, query=True) and not enable:
    #     cmds.warning('Closing nxt cmd port...')
    #     cmds.commandPort(name=address, cl=True)
    #     model = nxt.remote.nxt_socket.get_nxt_model()
    #     model.close(notify_server=True)


def create_remote_context(*args):
    t = 'maya' + cmds.about(version=True)
    if cmds.about(mac=True) or cmds.about(linux=True):
        partial_exe_path = 'bin/mayapy'
    elif cmds.about(win=True):
        partial_exe_path = 'bin/mayapy.exe'
    else:
        raise OSError('You are running an unsupported OS.')
    mayapy = os.path.join(os.environ['MAYA_LOCATION'], partial_exe_path)
    create_func = nxt_editor.main_window.MainWindow.create_remote_context
    create_func(place_holder_text=t, interpreter_exe=mayapy)


class NxtUiCmd(om.MPxCommand):
    cmd_name = "nxt_ui"

    @staticmethod
    def cmdCreator():
        return NxtUiCmd()

    def doIt(self, args):
        global __NXT_INSTANCE__
        os.environ[NXT_DCC_ENV_VAR] = 'maya'
        if args:
            string_args = []
            for arg in range(len(args)):
                string_args += [args.asString(arg)]
            if 'close' in string_args:
                if __NXT_INSTANCE__:
                    __NXT_INSTANCE__.close()
                return
        nxt_win = nxt_editor.main_window.MainWindow()
        if 'win32' in sys.platform:
            # gives nxt it's own entry on taskbar
            nxt_win.setWindowFlags(QtCore.Qt.Window)

        def log_callback(message, msg_type, data):
            formatting = {
                om.MCommandMessage.kWarning: '# Warning: {} #',
                om.MCommandMessage.kError: '# Error: {} #',
                om.MCommandMessage.kResult: '# Result: {} #'
            }
            if message.endswith('\n'):
                text = message[:-1]
            else:
                text = message
            text = formatting.get(msg_type, '{}').format(text)
            display_type = om.MCommandMessage.kDisplay
            if message.endswith('\n') or msg_type != display_type:
                text += '\n'
            nxt_win.output_log.write_raw.emit(text, time.time())
            model = nxt_win.model
            if model:
                model.process_events()
        cb_id = om.MCommandMessage.addCommandOutputCallback(log_callback, None)
        sj = cmds.scriptJob(e=["quitApplication", "cmds.nxt_ui('close')"],
                            protected=True)
        nxt_win.output_log.unwrap_std_streams()

        def remove_callback():
            om.MCommandMessage.removeCallback(cb_id)
            try:
                cmds.scriptJob(kill=sj, force=True)
            except RuntimeError:
                # During a real close it will try to kill the job while its
                # running. Maybe we should just block the signal?
                pass
        nxt_win.close_signal.connect(remove_callback)
        nxt_win.show()
        __NXT_INSTANCE__ = nxt_win


# PLUGIN BOILERPLATE #
def maya_useNewAPI(): pass


def initializePlugin(plugin):
    vendor = 'The nxt contributors'
    version = MAYA_PLUGIN_VERSION.VERSION
    pluginFn = om.MFnPlugin(plugin, vendor, version)
    # Commands
    # TODO promote to for loop if building multiple commands(same for uninit)
    try:
        pluginFn.registerCommand(NxtUiCmd.cmd_name, NxtUiCmd.cmdCreator)
    except Exception:
        logger.exception("Failed to register: {}".format(NxtUiCmd.cmd_name))
        raise
    # UI
    maya_window = mel.eval('$_=$gMainWindow')
    nxt_menu = cmds.menu('nxt', parent=maya_window, tearOff=True)
    CREATED_UI.append(nxt_menu)
    cmds.menuItem('Open Editor', command=cmds.nxt_ui, parent=nxt_menu)
    cmds.menuItem('Create Maya Context', command=create_remote_context,
                  parent=nxt_menu)
    # cmds.menuItem('Open Command Port', command=enable_cmd_port,
    #               parent=nxt_menu, checkBox=False)
    cmds.menuItem('About', command=about_menu, parent=nxt_menu)


def uninitializePlugin(plugin):
    # TODO: Long term we need to remove our self from the modules during the
    #  plugin unload.
    pluginFn = om.MFnPlugin(plugin)
    # Commands
    try:
        pluginFn.deregisterCommand(NxtUiCmd.cmd_name)
    except Exception:
        logger.exception("Failed to unregister: {}".format(NxtUiCmd.cmd_name))
        raise
    # UI
    for ui in CREATED_UI:
        cmds.deleteUI(ui, menu=True)



"""
Loosely based on the example addon from this repo:
https://github.com/robertguetzkow/blender-python-examples
"""
# Builtin
import os
import sys
import subprocess

# External
import bpy

try:
    # External
    from Qt import QtCore, QtWidgets
    # Internal
    from nxt_editor.constants import NXT_WEBSITE
    from nxt_editor.integration import blender
    nxt_installed = True
except ImportError:
    nxt_installed = False
    NXT_WEBSITE = 'https://nxt-dev.github.io/'

nxt_package_name = 'nxt-editor'

bl_info = {
    "name": "NXT Blender",
    "blender": (3, 4, 0),
    "version": (0, 3, 0),
    "location": "NXT > Open Editor",
    "wiki_url": "https://nxt-dev.github.io/",
    "tracker_url": "https://github.com/nxt-dev/nxt_editor/issues",
    "category": "nxt",
    "description": "NXT is a general purpose code compositor designed for "
                   "rigging, scene assembly, and automation. (This is an "
                   "experimental version of nxt_blender. Save "
                   "early, save often.)",
    "warning": "This addon requires installation of dependencies."
}

b_major, b_minor, b_patch = bpy.app.version
if b_major == 2:
    bl_info["blender"] = (2, 80, 0)
elif b_major != 3:
    raise RuntimeError('Unsupported major Blender version: {}'.format(b_major))


class BLENDER_PLUGIN_VERSION(object):
    plugin_v_data = {'MAJOR': bl_info["version"][0],
                     'MINOR': bl_info["version"][1],
                     'PATCH': bl_info["version"][2]}
    MAJOR = plugin_v_data['MAJOR']
    MINOR = plugin_v_data['MINOR']
    PATCH = plugin_v_data['PATCH']
    VERSION_TUPLE = (MAJOR, MINOR, PATCH)
    VERSION_STR = '.'.join(str(v) for v in VERSION_TUPLE)
    VERSION = VERSION_STR


class CreateBlenderContext(bpy.types.Operator):
    bl_label = "Create Remote Blender NXT Context"
    bl_idname = "nxt.create_blender_context"

    def execute(self, context):
        global nxt_installed
        if nxt_installed:
            b = blender.__NXT_INTEGRATION__
            if not b:
                b = blender.Blender.launch_nxt()
            b.create_context()
        else:
            show_dependency_warning()
        return {'FINISHED'}


class OpenNxtEditor(bpy.types.Operator):
    bl_label = "Open NXT Editor"
    bl_idname = "nxt.nxt_editor"

    def execute(self, context):
        global nxt_installed
        if nxt_installed:
            blender.Blender.launch_nxt()
        else:
            show_dependency_warning()
        return {'FINISHED'}


class AboutNxt(bpy.types.Operator):
    bl_label = "Update NXT"
    bl_idname = "nxt.nxt_about"

    def execute(self, context):
        import webbrowser
        webbrowser.open_new(NXT_WEBSITE)
        return {'FINISHED'}


class TOPBAR_MT_nxt(bpy.types.Menu):
    bl_label = "NXT"

    def draw(self, context):
        layout = self.layout
        layout.operator("nxt.nxt_editor", text="Open Editor")
        layout.separator()
        layout.operator("nxt.nxt_update_dependencies",
                        text="Update NXT (Requires Blender Restart)")
        layout.separator()
        layout.operator('nxt.create_blender_context', text='Create Blender '
                                                           'Context')
        layout.separator()
        layout.operator("nxt.nxt_about", text="About")

    def menu_draw(self, context):
        self.layout.menu("TOPBAR_MT_nxt")


class NxtInstallDependencies(bpy.types.Operator):
    bl_idname = 'nxt.nxt_install_dependencies'
    bl_label = "Install NXT dependencies (Blender requires elevated permissions)"
    bl_description = ("Downloads and installs the required python packages "
                      "for NXT. Internet connection is required. "
                      "Blender may have to be started with elevated "
                      "permissions in order to install the package. "
                      "Alternatively you can pip install nxt-editor into your "
                      "Blender Python environment.")
    bl_options = {"REGISTER", "INTERNAL"}

    @classmethod
    def poll(cls, context):
        global nxt_installed
        return not nxt_installed

    def execute(self, context):
        environ_copy = dict(os.environ)
        environ_copy["PYTHONNOUSERSITE"] = "1"
        pkg = 'nxt-editor'
        if b_major == 2:
            exe = bpy.app.binary_path_python
        else:
            exe = sys.executable
        try:
            subprocess.run([exe, "-m", "pip", "install", pkg],
                           check=True, env=environ_copy)
        except subprocess.CalledProcessError as e:
            self.report({"ERROR"}, str(e))
            return {"CANCELLED"}
        msg = 'Please restart Blender to finish installing NXT.'
        self.report({"INFO"}, msg)
        show_message(msg, "Installed dependencies!")
        return {"FINISHED"}


class NxtUpdateDependencies(bpy.types.Operator):
    bl_idname = 'nxt.nxt_update_dependencies'
    bl_label = "Update NXT dependencies"
    bl_description = ("Downloads and updates the required python packages "
                      "for NXT. Internet connection is required. "
                      "Blender may have to be started with elevated "
                      "permissions in order to install the package. "
                      "Alternatively you can pip install -U nxt-editor into "
                      "your Blender Python environment.")
    bl_options = {"REGISTER", "INTERNAL"}

    @classmethod
    def poll(cls, context):
        global nxt_installed
        return nxt_installed

    def execute(self, context):
        try:
            blender.Blender._update_package('nxt-editor')
        except subprocess.CalledProcessError as e:
            self.report({"ERROR"}, str(e))
            return {"CANCELLED"}
        self.report({"INFO"}, 'Please restart Blender to '
                              'finish updating NXT.')
        return {"FINISHED"}


class NxtUninstallDependencies(bpy.types.Operator):
    bl_idname = 'nxt.nxt_uninstall_dependencies'
    bl_label = "Uninstall NXT dependencies"
    bl_description = ("Uninstalls the NXT Python packages. "
                      "Blender may have to be started with elevated "
                      "permissions in order to install the package. "
                      "Alternatively you can pip uninstall nxt-editor from "
                      "your Blender Python environment.")
    bl_options = {"REGISTER", "INTERNAL"}

    @classmethod
    def poll(cls, context):
        global nxt_installed
        return nxt_installed

    def execute(self, context):
        try:
            blender.Blender().uninstall()
        except subprocess.CalledProcessError as e:
            self.report({"ERROR"}, str(e))
            return {"CANCELLED"}
        self.report({"INFO"}, 'Please restart Blender to '
                              'finish uninstalling NXT dependencies.')
        return {"FINISHED"}


class NxtDependenciesPreferences(bpy.types.AddonPreferences):
    bl_idname = __name__

    def draw(self, context):
        layout = self.layout
        layout.operator(NxtInstallDependencies.bl_idname, icon="PLUGIN")
        layout.operator(NxtUpdateDependencies.bl_idname, icon="SCRIPT")
        layout.operator(NxtUninstallDependencies.bl_idname, icon="PANEL_CLOSE")


def show_dependency_warning():

    def draw(self, context):
        layout = self.layout
        lines = [
            f"Please install the missing dependencies for the NXT add-on.",
            "1. Open the preferences (Edit > Preferences > Add-ons).",
            f"2. Search for the \"{bl_info.get('name')}\" add-on.",
            "3. Open the details section of the add-on.",
            f"4. Click on the \"{NxtInstallDependencies.bl_label}\" button.",
            "This will download and install the missing Python packages. "
            "You man need to start Blender with elevated permissions",
            f"Alternatively you can pip install \"{nxt_package_name}\" into "
            f"your Blender Python environment."
        ]

        for line in lines:
            layout.label(text=line)
    bpy.context.window_manager.popup_menu(draw, title='NXT Warning!',
                                          icon="ERROR")


def show_message(message, title, icon='INFO'):

    def draw(self, *args):
        self.layout.label(text=message)

    bpy.context.window_manager.popup_menu(draw, title=title, icon=icon)


nxt_operators = (TOPBAR_MT_nxt, OpenNxtEditor, NxtUpdateDependencies,
                 NxtUninstallDependencies, NxtDependenciesPreferences,
                 NxtInstallDependencies, CreateBlenderContext)


def register():
    global nxt_installed
    for cls in nxt_operators:
        bpy.utils.register_class(cls)
    bpy.utils.register_class(AboutNxt)
    bpy.types.TOPBAR_MT_editor_menus.append(TOPBAR_MT_nxt.menu_draw)


def unregister():
    try:
        if blender.__NXT_INTEGRATION__:
            blender.__NXT_INTEGRATION__.quit_nxt()
    except NameError:
        pass
    bpy.types.TOPBAR_MT_editor_menus.remove(TOPBAR_MT_nxt.menu_draw)
    for cls in nxt_operators:
        bpy.utils.unregister_class(cls)
    bpy.utils.unregister_class(AboutNxt)


if __name__ == "__main__":
    register()



# Built-in
import logging

# External
from Qt import QtWidgets, QtGui, QtCore

# Internal
import nxt_editor
from nxt_editor.dockwidgets.dock_widget_base import DockWidgetBase
from nxt_editor import colors, dialogs

logger = logging.getLogger(nxt_editor.LOGGER_NAME)
TOOLTIP_INFO = ('<p style="font-size:12px;color:white;">'
                '<h3>How to:</h3>'
                '<ul>'
                '<li>'
                'To edit a hotkey double click a cell in the <b>Shortcut</b> '
                'column and press a key combination. </li><li>'
                'When you release the key(s) the hotkey is temporally stored.'
                '</li><li>'
                'When you are ready to apply your changes click the <b>Save</b>'
                ' changes button at the bottom of the widget. </li><li>'
                'To revert or remove a hotkey, right click the shortcut cell.'
                '</li></ul>'
                '<h3>Note:</h3>'
                'Actions with <span style="color:#dfdf16;">yellow</span> text '
                'are global and cannot have a shortcut that conflicts with any '
                'other shortcut.</p>')

TOOLTIP_STYLE = '''QToolTip {
                    font-family: Roboto Mono;
                    background-color: #3E3E3E;
                    border: 1px solid #232323;
                }'''

TABLE_STYLE = '''QTableView {
                    font-family: Roboto Mono;
                }'''


class HotkeyEditor(DockWidgetBase):
    savable = QtCore.Signal(bool)
    refresh = QtCore.Signal()

    def __init__(self, parent):
        super(HotkeyEditor, self).__init__('Hotkey Editor', parent,
                                           minimum_height=250)
        self.setWindowFlags(QtCore.Qt.Tool)
        self.main_window = parent
        self.main_widget = QtWidgets.QWidget(parent=self)
        self.setWidget(self.main_widget)

        self.layout = QtWidgets.QHBoxLayout()
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.setSpacing(0)
        self.main_widget.setLayout(self.layout)

        self.background_frame = QtWidgets.QFrame(self)
        self.layout.addWidget(self.background_frame)

        self.main_frame = QtWidgets.QFrame(self)
        self.layout.addWidget(self.main_frame)
        self.hotkey_layout = QtWidgets.QVBoxLayout()
        self.hotkey_layout.setContentsMargins(8, 0, 8, 0)
        self.hotkey_layout.setSpacing(0)
        self.main_frame.setLayout(self.hotkey_layout)
        self.hotkey_table_view = HotkeyView(self)
        self.hotkey_table_view.setToolTip(TOOLTIP_INFO)
        self.hotkey_table_view.horizontalHeader().setSectionsMovable(False)
        self.hotkey_table_view.horizontalHeader().setStretchLastSection(True)
        double_click = QtWidgets.QAbstractItemView.DoubleClicked
        self.hotkey_table_view.setEditTriggers(double_click)
        single_select = QtWidgets.QAbstractItemView.SingleSelection
        self.hotkey_table_view.setSelectionMode(single_select)
        self.hotkey_table_view.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.hotkey_table_view.customContextMenuRequested.connect(self.context_menu)

        self.hotkey_layout.addWidget(self.hotkey_table_view)
        self.hotkey_model = HotkeyModel(self, view=self.hotkey_table_view)
        self.hotkey_table_view.setModel(self.hotkey_model)
        self.hb_buttons = QtWidgets.QHBoxLayout()
        self.hotkey_layout.addLayout(self.hb_buttons)
        self.btn_discard = QtWidgets.QPushButton('Discard changes')
        self.btn_discard.setEnabled(False)

        def _discard():
            self.hotkey_model.discard_changes(warn=False)
        self.btn_discard.clicked.connect(_discard)
        self.savable.connect(self.btn_discard.setEnabled)
        self.hb_buttons.addWidget(self.btn_discard)
        self.btn_save = QtWidgets.QPushButton('Save & Apply changes')
        self.btn_save.setEnabled(False)
        self.btn_save.clicked.connect(self.hotkey_model.save)
        self.savable.connect(self.btn_save.setEnabled)
        self.hb_buttons.addWidget(self.btn_save)
        self.refresh.connect(self.hotkey_model.discard_changes)
        self.resize(self.hotkey_table_view.width()*8, self.minimumHeight())
        self.refresh.emit()

    def context_menu(self, pos):
        index = self.hotkey_table_view.indexAt(pos)
        self.hotkey_table_view.closePersistentEditor(index)
        if index.column() != len(self.hotkey_model.header_names)-1:
            return
        menu = QtWidgets.QMenu(self)
        menu.addAction('Revert Hotkey', self.revert_hotkey)
        menu.addAction('Remove Hotkey', self.remove_hotkey)
        menu.popup(QtGui.QCursor.pos())

    def revert_hotkey(self):
        index = self.hotkey_table_view.selectionModel().selectedIndexes()[0]
        self.hotkey_model.revert_hotkey(index)

    def remove_hotkey(self):
        index = self.hotkey_table_view.selectionModel().selectedIndexes()[0]
        self.hotkey_model.remove_hotkey(index)

    def closeEvent(self, event):
        allow = self.hotkey_model.discard_changes()
        if allow:
            return super(HotkeyEditor, self).closeEvent(event)
        return event.ignore()


class HotkeyModel(QtCore.QAbstractTableModel):
    def __init__(self, parent, view, headers=None):
        super(HotkeyModel, self).__init__()
        self.parent = parent
        self.header_names = headers or ['Name',
                                        'What\'s This', 'Tool Tip', 'Hotkey']
        self.view = view
        self.node_attr_names = list()
        self.attr_data = list()
        self.horizontal_header = self.view.horizontalHeader()
        self.state = None
        self.delegate = KeySequenceDelegate(parent=self)
        # set default data
        self._data = []
        self.user_changes = {}
        self.actions = {}
        self.cast_mode = QtGui.QKeySequence.PortableText
        self.protected_actions = []
        self.protected_shortcuts = []

    def discard_changes(self, warn=True):
        """Removes any changes made by the user.
        :param warn: If True a warning dialog will be shown allowing the user
        to cancel the discard.
        :return: False if user wants to cancel discard, True if user saved or
        discarded changes
        """
        if warn is True and self.user_changes.keys():
            info = 'Would you like to save your shortcuts?'
            resp = dialogs.UnsavedChangesMessage.save_before_close(info=info)
            save = dialogs.UnsavedChangesMessage.Save
            cancel = dialogs.UnsavedChangesMessage.Cancel
            if resp == cancel:
                return False
            elif resp == save:
                self.save()
        self.user_changes = {}
        self.update_data()
        self.parent.savable.emit(False)
        return True

    def update_data(self):
        nxt_hotkeys = self.parent.main_window.get_hotkey_map()
        self.clear()
        self.actions = {}
        self.protected_actions = []
        self.protected_shortcuts = []
        for section, widget_actions in nxt_hotkeys.items():
            # Div rows have a None object as their last item.
            div_row = [section, 'What\'s this', 'Tooltip', 'Shortcut', None]
            self._data.append(div_row)
            for action_data in widget_actions:
                action = action_data[4]
                self._data += [action_data]
                self.actions[action] = action.isEnabled()
                if action.shortcutContext() == QtCore.Qt.WindowShortcut:
                    self.protected_actions += [action]
                    self.protected_shortcuts += [action_data[3]]
        if nxt_hotkeys:
            fixed = QtWidgets.QHeaderView.Fixed
            self.horizontal_header.setSectionResizeMode(0, fixed)
            self.horizontal_header.setDefaultSectionSize(200)
            self.view.setItemDelegateForColumn(self.columnCount()-1,
                                               self.delegate)
            self.view.resizeRowsToContents()

    def update_protected_shortcuts(self, action, old_shortcut, new_shortcut):
        if action in self.protected_actions:
            if old_shortcut in self.protected_shortcuts:
                self.protected_shortcuts.remove(old_shortcut)
            if new_shortcut:
                self.protected_shortcuts += [new_shortcut]

    def clear(self):
        self.beginResetModel()
        self._data = []
        self.endResetModel()

    def save(self):
        for i, data in self.user_changes.items():
            shortcut, action = data
            action.setShortcut(shortcut, user_override=True)
        self.user_changes = {}
        self.update_data()
        self.parent.savable.emit(False)

    def revert_hotkey(self, index):
        row = index.row()
        action = self._data[row][self.columnCount()]
        old_value = self._data[row][3]
        value = action.default_shortcut
        if value == old_value:
            return
        if not self.valid_shortcut(action, value):
            self.show_invalid_message(value)
            return
        self.update_protected_shortcuts(action, old_value, value)
        self.setData(index, value, QtCore.Qt.EditRole)

    def valid_shortcut(self, action, shortcut):
        valid = True
        if action in self.protected_actions:
            if shortcut in self.all_shortcuts:
                valid = False
        elif shortcut in self.protected_shortcuts:
            valid = False
        return valid

    @staticmethod
    def show_invalid_message(shortcut):
        message = 'Invalid key sequence: {} \nIt would create a ' \
                  'conflict for a global shortcut'.format(shortcut)
        dialogs.NxtWarningDialog.show_message('Invalid shortcut!', message)

    def remove_hotkey(self, index):
        row = index.row()
        action = self._data[row][self.columnCount()]
        old_value = self._data[row][3]
        if not old_value:
            return
        self.update_protected_shortcuts(action, old_value, None)
        self.setData(index, None, QtCore.Qt.EditRole)

    def disable_actions(self):
        for action in self.actions.keys():
            action.setEnabled(False)

    def enable_actions(self):
        for action in self.actions.keys():
            action.setEnabled(self.actions[action])

    def data(self, index, role=None):
        if not index.isValid:
            return None
        row = index.row()
        column = index.column()
        if role == QtCore.Qt.DisplayRole:
            return self._data[row][column]
        elif role == QtCore.Qt.EditRole:
            # Disable all the actions so the user input doesn't trigger one
            # I tried an event filter but it didn't work on application
            # shortcuts like Ctrl+S
            self.disable_actions()
            return self._data[row][column]
        elif role == QtCore.Qt.ForegroundRole:
            action = self._data[row][self.columnCount()]
            if action and action.shortcutContext() == QtCore.Qt.WindowShortcut:
                color = colors.IMPORTANT
                return QtGui.QBrush(color)
        elif role == QtCore.Qt.BackgroundRole:
            if row in self.user_changes.keys():
                color = colors.UNSAVED
                return QtGui.QBrush(color, QtCore.Qt.BDiagPattern)
            actions = self._data[row][self.columnCount()]
            if not actions:
                # TODO: Better style management
                color = QtGui.QColor(QtCore.Qt.darkGray).darker(250)
                return QtGui.QBrush(color)
        elif role == QtCore.Qt.FontRole:
            action = self._data[row][self.columnCount()]
            if action:
                default_shortcut = action.default_shortcut
                if isinstance(default_shortcut, QtGui.QKeySequence):
                    default_shortcut = default_shortcut.toString()
                shortcut = self._data[row][self.columnCount()-1] or None
                if shortcut != default_shortcut:
                    font = QtGui.QFont()
                    font.setItalic(True)
                    return font
            else:
                font = QtGui.QFont()
                font.setBold(True)
                font.setPointSizeF(font.pointSize()*1.1)
                font.setItalic(True)
                return font
        elif not role:
            return self._data[row][column]

    def setData(self, index, value, role):
        value_set = False
        if not index.isValid:
            return value_set
        row = index.row()
        column = index.column()
        # Reset the actions to their previous enabled state

        if role == QtCore.Qt.EditRole and column == self.columnCount()-1:
            if value != self._data[row][column]:
                self._data[row][column] = value
                action = self._data[row][column+1]
                # The last visible column holds the shortcut the column
                # after this holds the NxtAction object actions are always
                # in a list
                self.user_changes[row] = [value, action]
                value_set = True
            if row in list(self.user_changes.keys()):
                action = self.user_changes[row][1]
                if value == action.shortcut().toString(self.cast_mode):
                    self.user_changes.pop(row)
            self.enable_actions()
        if self.user_changes.keys():
            self.parent.savable.emit(True)
        else:
            self.parent.savable.emit(False)
        self.view.update()
        return value_set

    def flags(self, index):
        column = index.column()
        if column == self.columnCount()-1:
            return QtCore.Qt.ItemIsEnabled | \
                   QtCore.Qt.ItemIsSelectable | \
                   QtCore.Qt.ItemIsEditable
        else:
            return QtCore.Qt.NoItemFlags

    def headerData(self, section, orientation, role):
        if orientation is QtCore.Qt.Horizontal:
            if role is QtCore.Qt.DisplayRole:
                return self.header_names[section]

    def rowCount(self, parent):
        return len(self._data)

    def columnCount(self, *args):
        return len(self.header_names)

    @property
    def all_shortcuts(self):
        shortcuts = []
        for row in self._data:
            if row[self.columnCount()]:
                shortcuts += [row[3]]
        return shortcuts


class HotkeyView(QtWidgets.QTableView):
    def __init__(self, parent):
        super(HotkeyView, self).__init__(parent=parent)
        self.hotkey_editor = parent
        style = self.parent().styleSheet() + TOOLTIP_STYLE + TABLE_STYLE
        self.setStyleSheet(style)


class KeySequenceDelegate(QtWidgets.QStyledItemDelegate):

    def __init__(self, parent):
        super(KeySequenceDelegate, self).__init__(parent=parent)
        self.hotkey_model = parent

    def createEditor(self, parent, option, index):
        last_col = self.hotkey_model.columnCount()
        action = self.hotkey_model._data[index.row()][last_col]
        valid_cell = bool(action)
        if not valid_cell:
            # Don't allow users to edit rows that don't have actions in them
            return
        line_edit = KeySequenceEdit(parent, index.data(), action,
                                    self.hotkey_model)
        return line_edit


class KeySequenceEdit(QtWidgets.QLineEdit):
    """
    Based on https://gist.github.com/blink1073/946df268c3685a3f443e
    """

    def __init__(self, parent, key_sequence, action, hotkey_model):
        super(KeySequenceEdit, self).__init__(parent)
        self.action = action
        self.parent = parent
        self.hotkey_model = hotkey_model
        self.setText(key_sequence)
        self.press_count = 0
        self.keys = set()
        self.modifiers = ['Meta', 'Ctrl', 'Alt', 'Shift']
        self.input_text = key_sequence
        self.installEventFilter(self)
        self.setFocusPolicy(QtCore.Qt.StrongFocus)
        self.cast_mode = QtGui.QKeySequence.PortableText
        self.setContextMenuPolicy(QtCore.Qt.NoContextMenu)

    def keyPressEvent(self, event):
        if event.isAutoRepeat():
            return
        key = event.key()
        self.press_count += 1

        if key == QtCore.Qt.Key_unknown:
            logger.error("Unknown key from a macro probably")
            return
        event_modifiers = event.modifiers()
        modifier_key_map = {QtCore.Qt.Key_Control: QtCore.Qt.CTRL,
                         QtCore.Qt.Key_Shift: QtCore.Qt.SHIFT,
                         QtCore.Qt.Key_Alt: QtCore.Qt.ALT,
                         QtCore.Qt.Key_Meta: QtCore.Qt.META}
        if key in list(modifier_key_map.keys()) and not len(self.keys):
            shift_held = bool(event_modifiers & QtCore.Qt.ShiftModifier)
            ctrl_held = bool(event_modifiers & QtCore.Qt.ControlModifier)
            alt_held = bool(event_modifiers & QtCore.Qt.AltModifier)
            meta_held = bool(event_modifiers & QtCore.Qt.MetaModifier)
            held_modifiers = [meta_held, ctrl_held, alt_held, shift_held]
            if sum(held_modifiers) > 1:
                text = ''
                for modifier, held in zip(self.modifiers, held_modifiers):
                    if held:
                        text += '{}+'.format(modifier)
                text = text[:-1]
            else:
                key = modifier_key_map[key]
                keySequence = QtGui.QKeySequence(key)
                text = keySequence.toString()
            if not isinstance(text, str):
                text = text.decode()
            self.setText(text)
            return
        else:
            if event_modifiers & QtCore.Qt.ShiftModifier:
                key += QtCore.Qt.SHIFT
            if event_modifiers & QtCore.Qt.ControlModifier:
                key += QtCore.Qt.CTRL
                self.press_count -= 1
            if event_modifiers & QtCore.Qt.AltModifier:
                key += QtCore.Qt.ALT
            if event_modifiers & QtCore.Qt.MetaModifier:
                key += QtCore.Qt.META

            self.keys.add(key)
        if len(self.keys) > 4:
            logger.error("Too many keys, max 4!")
            text = 'Too many keys, max 4!'
            self.keys = set()
            self.press_count = 0
        else:
            keySequence = QtGui.QKeySequence(*self.keys)
            text = keySequence.toString(self.cast_mode)
            for modifier in self.modifiers:
                if text.count(modifier) > 1:
                    text = 'Invalid key combo'
                    self.keys = set()
                    self.press_count = 0
        self.setText(text)
        event.accept()

    def keyReleaseEvent(self, event):
        self.press_count = 0
        self.keys = set()
        key_seq = QtWidgets.QShortcut(QtGui.QKeySequence(self.text()), self)
        keys = key_seq.key().toString()
        valid = self.hotkey_model.valid_shortcut(self.action, keys)
        if keys != self.text():
            logger.error('Invalid key sequence! \'{}\''.format(self.text()))
            self.setText(self.input_text)
        elif not valid:
            self.setText(self.input_text)
            self.hotkey_model.show_invalid_message(keys)
            return
        self.hotkey_model.update_protected_shortcuts(self.action,
                                                     self.input_text, keys)
        self.clearFocus()
        event.accept()

    def focusOutEvent(self, event):
        self.hotkey_model.enable_actions()
        super(KeySequenceEdit, self).focusOutEvent(event)



import os

from Qt import QtCore, QtWidgets, QtGui


class SearchLineEdit(QtWidgets.QLineEdit):

    file_chosen = QtCore.Signal()

    def __init__(self, parent=None):
        super(SearchLineEdit, self).__init__(parent)
        completer = QtWidgets.QCompleter(parent=self)
        model = QtWidgets.QFileSystemModel(completer)
        model.setNameFilters(['*.nxt', '*.nxtb'])
        model.setNameFilterDisables(False)
        completer.setModel(model)
        completer.setCaseSensitivity(QtCore.Qt.CaseInsensitive)
        completer.setCompletionMode(completer.UnfilteredPopupCompletion)
        completer.popup().setStyleSheet(parent.parent().stylesheet)
        self.returnPressed.connect(self.file_chosen.emit)
        self.setCompleter(completer)
        self.setFocusPolicy(QtCore.Qt.FocusPolicy.StrongFocus)

    def focusOutEvent(self, event):
        super(SearchLineEdit, self).focusOutEvent(event)
        self.parent().close()

    def keyPressEvent(self, event):
        super(SearchLineEdit, self).keyPressEvent(event)
        if event.key() == QtCore.Qt.Key_Escape:
            self.setText('')
            self.file_chosen.emit()
            return
        if event.key() == QtCore.Qt.Key_Return:
            return
        if os.path.isdir(self.text()) and self.text()[-1] == os.path.sep:
            QtCore.QTimer.singleShot(0, self.completer().complete)

    def setText(self, text):
        super(SearchLineEdit, self).setText(text)
        self.completer().model().setRootPath(text)


class Searcher(QtWidgets.QDialog):
    def __init__(self, parent, width, default=''):
        super(Searcher, self).__init__(parent, QtCore.Qt.FramelessWindowHint)
        self.setAttribute(QtCore.Qt.WA_TranslucentBackground)
        self.setFixedHeight(64)
        self.setFixedWidth(width)
        self.layout = QtWidgets.QVBoxLayout(self)
        self.line_edit = SearchLineEdit(self)
        default = os.path.normcase(default)
        self.line_edit.setText(default)
        self.line_edit.setStyleSheet(self.parent().stylesheet)
        self.line_edit.file_chosen.connect(self.close)
        self.layout.addWidget(self.line_edit)

    @classmethod
    def get_open_file_path(cls, *args, **kwargs):
        inst = cls(*args, **kwargs)
        inst.exec_()
        return inst.line_edit.text()



from nxt_editor.dockwidgets.dock_widget_base import DockWidgetBase
from nxt_editor.dockwidgets.build_view import BuildView
from nxt_editor.dockwidgets.code_editor import CodeEditor
from nxt_editor.dockwidgets.history_view import HistoryView
from nxt_editor.dockwidgets.hotkey_editor import HotkeyEditor
from nxt_editor.dockwidgets.layer_manager import LayerManager
from nxt_editor.dockwidgets.output_log import OutputLog
from nxt_editor.dockwidgets.property_editor import PropertyEditor
from nxt_editor.dockwidgets.widget_builder import WidgetBuilder
from nxt_editor.dockwidgets.find_rep import FindRepDockWidget



# Built-in
import logging
import ast
try:
    from collections.abc import Iterable
except ImportError:
    from collections import Iterable

# External
from Qt import QtWidgets
from Qt import QtGui
from Qt import QtCore
# Maya fix
try:
    QtCore.QStringListModel
except AttributeError:
    del QtCore
    from PySide2 import QtCore

# Internal
import nxt_editor
from nxt_editor import user_dir
from nxt_editor.dockwidgets.dock_widget_base import DockWidgetBase
from nxt import DATA_STATE, nxt_path
from nxt import nxt_node

logger = logging.getLogger(nxt_editor.LOGGER_NAME)

WINDOW_ATTR = '_widget_window'
WIDGET_TYPES = ['window', 'tab', 'panel', 'gridLayout', 'button', 'menuItem',
                'dropDownMenu', 'checkbox']
RECOMP_PREF = user_dir.USER_PREF.RECOMP_PREF


class WidgetBuilder(DockWidgetBase):

    WINDOW_TITLE_ATTR = 'window_title'
    WINDOW_COLOR_ATTR = 'background_color'

    def __init__(self, graph_model=None, title='Workflow Tools', parent=None,
                 minimum_width=50, minimum_height=50):
        super(WidgetBuilder, self).__init__(title=title,
                                            parent=parent,
                                            minimum_width=minimum_width,
                                            minimum_height=minimum_height)

        self.updating = False
        self.setObjectName('Workflow Tools')
        self.default_title = title

        # local attributes
        self.main_window = parent
        self.stage_model = graph_model
        self.window_node_path = None
        self.window_title = None

        # state attributes
        self.widget_state_data = {}
        self.tab_widgets = []
        self.scroll_pos = 0

        # main layout
        self.main = QtWidgets.QWidget(parent=self)
        self.setWidget(self.main)

        self.layout = QtWidgets.QVBoxLayout()
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.setSpacing(0)
        self.main.setLayout(self.layout)

        # background frame
        self.background_frame = QtWidgets.QFrame(self)
        self.background_frame.setStyleSheet(
            'QFrame {background-color: #3E3E3E; border-radius: 0px;}')
        self.layout.addWidget(self.background_frame)

        self.background_layout = QtWidgets.QVBoxLayout()
        self.background_layout.setContentsMargins(4, 4, 4, 4)
        self.background_frame.setLayout(self.background_layout)

        # window
        self.window_frame = QtWidgets.QFrame(self)
        self.background_layout.addWidget(self.window_frame)

        self.window_layout = QtWidgets.QVBoxLayout()
        self.window_layout.setContentsMargins(4, 4, 0, 4)
        self.window_layout.setSpacing(0)
        self.window_frame.setLayout(self.window_layout)

        # scroll area
        self.scroll_area = QtWidgets.QScrollArea(self)
        self.scroll_area.setWidgetResizable(True)
        self.window_layout.addWidget(self.scroll_area)

        self.scroll_widget = QtWidgets.QWidget(self.window_frame)
        self.scroll_area.setWidget(self.scroll_widget)

        self.scroll_layout = QtWidgets.QVBoxLayout(self.scroll_widget)
        self.scroll_layout.setContentsMargins(0, 0, 4, 0)
        self.scroll_layout.setSpacing(4)
        self.scroll_layout.setAlignment(QtCore.Qt.AlignTop)
        self.scroll_widget.setLayout(self.scroll_layout)
        # Context menu
        self.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self.context_menu)

        # update
        self.update_window()

    def context_menu(self):
        menu = QtWidgets.QMenu(self)
        refresh_action = menu.addAction('Refresh')
        refresh_action.triggered.connect(self.update_window)
        menu.addAction(self.main_window.execute_actions.wt_recomp_action)
        menu.popup(QtGui.QCursor.pos())

    def show(self):
        super(WidgetBuilder, self).show()
        self.update_window()

    def set_stage_model(self, stage_model):
        super(WidgetBuilder, self).set_stage_model(stage_model)
        if self.stage_model:
            self.tab_widgets = []
            self.update_window()

    def set_stage_model_connections(self, model, connect):
        self.model_signal_connections = [
            (model.nodes_changed, self.update_window),
            (model.attrs_changed, self.update_window),
            (model.comp_layer_changed, self.update_window),
            (model.executing_changed, self.on_executing_changed)
        ]
        super(WidgetBuilder, self).set_stage_model_connections(model, connect)

    def on_executing_changed(self, state):
        self.window_frame.setVisible(not state)

    def on_stage_model_destroyed(self):
        super(WidgetBuilder, self).on_stage_model_destroyed()
        self.tab_widgets = []
        self.window_frame.hide()

    def build_widgets(self, node_path, layout):
        child_paths = self.stage_model.get_children(
            node_path=node_path,
            layer=self.stage_model.comp_layer,
            ordered=True)
        for child_path in child_paths:
            if not self.stage_model.get_node_enabled(child_path):
                self.build_widgets(node_path=child_path, layout=layout)
                continue

            widget_type = get_widget_type(child_path, self.stage_model)
            # tabs
            if widget_type == 'tab':
                # get existing tab widget
                count = layout.count()
                tab_widget = None
                if count:
                    for child_index in range(count):
                        child_item = layout.itemAt(child_index)
                        if not child_item:
                            continue
                        child_widget = child_item.widget()
                        if child_widget.__class__.__name__ == 'QTabWidget':
                            tab_widget = child_widget

                # create new tab widget
                if not tab_widget:
                    tab_widget = QtWidgets.QTabWidget(self)
                    tab_widget.setContextMenuPolicy(
                        QtCore.Qt.CustomContextMenu)
                    tab_widget.customContextMenuRequested.connect(
                        lambda: self.tab_context_menu(tab_widget))
                    tab_widget.currentChanged.connect(
                        lambda: self.set_tab_style(tab_widget))
                    tab_widget.currentChanged.connect(
                        lambda: self.set_tab_index(tab_widget))
                    layout.addWidget(tab_widget)
                    tab_widget_path = 'tab_widget_' + str(len(self.tab_widgets))
                    tab_widget.tab_widget_path = tab_widget_path
                else:
                    tab_widget_path = tab_widget.tab_widget_path

                # create tab
                tab_widget.blockSignals(True)
                widget = Tab(node_path=child_path,
                             tab_parent=tab_widget,
                             parent=self)
                tab_widget.blockSignals(False)
                self.build_widgets(node_path=child_path, layout=widget.layout)

                # set tab state
                data = self.widget_state_data.get(tab_widget_path)
                if data:
                    if data['widget_type'] == widget_type:
                        state = data.get('state', 0)
                        tab_widget.setCurrentIndex(state)
                    else:
                        data['widget_type'] = widget_type
                        data['state'] = tab_widget.currentIndex()
                    data['widget'] = self
                else:
                    data = {'widget_type': widget_type,
                            'widget': tab_widget,
                            'state': tab_widget.currentIndex()}
                    self.widget_state_data[tab_widget_path] = data
                self.set_tab_style(tab_widget)

            # panels
            elif widget_type == 'panel':
                widget = Panel(node_path=child_path, parent=self)
                layout.addWidget(widget)
                self.build_widgets(node_path=child_path, layout=widget.layout)

            # grid layouts
            elif widget_type == 'gridLayout':
                widget = GridLayout(node_path=child_path, parent=self)
                layout.addWidget(widget)
                self.build_widgets(node_path=child_path, layout=widget.layout)

            # buttons
            elif widget_type == 'button':
                widget = Button(node_path=child_path, parent=self)
                layout.addWidget(widget)
                self.build_widgets(node_path=child_path, layout=layout)

            # checkboxes
            elif widget_type == 'checkbox':
                widget = CheckBox(node_path=child_path, parent=self)
                layout.addWidget(widget)

            # drop-down menus
            elif widget_type == 'dropDownMenu':
                widget = DropDownMenu(node_path=child_path, parent=self)
                layout.addWidget(widget)

            # unrecognized node
            else:
                self.build_widgets(node_path=child_path, layout=layout)
            self.stage_model.process_events()  # Visually update

    def get_window_node_path(self):
        if not self.stage_model:
            return
        if not self.stage_model.comp_layer:
            return

        all_node_paths = self.stage_model.comp_layer.descendants()
        window_nodes = []
        for node_path in all_node_paths:
            local_attrs = self.stage_model.get_node_local_attr_names(
                node_path,
                self.stage_model.comp_layer)
            if WINDOW_ATTR in local_attrs:
                window_nodes.append(node_path)

        window_node = None
        for node in window_nodes:
            value = self.stage_model.get_node_attr_value(
                node_path=node,
                attr_name=WINDOW_ATTR,
                layer=self.stage_model.comp_layer,
                data_state=DATA_STATE.RESOLVED)
            if value == 'True':
                window_node = node
                break

        return window_node

    def get_window_title(self):
        if self.window_node_path:
            title = self.stage_model.get_node_attr_value(
                        node_path=self.window_node_path,
                        attr_name=self.WINDOW_TITLE_ATTR,
                        layer=None)
            return title or None

    def update_window(self, changed_paths=None):
        if self.updating:
            return
        if not self.isVisible():
            return

        self.window_node_path = self.get_window_node_path()
        if not self.window_node_path:
            self.window_frame.hide()
            return

        update = False if changed_paths else True
        if not isinstance(changed_paths, Iterable):
            changed_paths = []
        for path in changed_paths:
            node_path, _ = nxt_path.path_attr_partition(path)
            is_widget = get_widget_type(node_path, self.stage_model)
            if is_widget:
                desc = self.stage_model.get_descendants(self.window_node_path)
                if node_path in desc:
                    update = True
                    break
            else:
                if not self.stage_model.node_exists(node_path):
                    update = True
                    break
        if not update:
            return
        self.updating = True
        # window title
        title = self.get_window_title()
        self.setWindowTitle(title or self.default_title)

        # remove existing widgets
        while self.scroll_layout.count():
            child = self.scroll_layout.itemAt(0)
            if child.widget():
                c = child.widget()
                c.setParent(None)
                c.deleteLater()
                del c

        # build widgets
        self.build_widgets(self.window_node_path, self.scroll_layout)
        self.set_window_style()
        self.window_frame.show()

        # add context menu
        ContextMenu(stage_model=self.stage_model,
                    node_path=self.window_node_path,
                    widget=self.window_frame,
                    items=None,
                    parent=self)
        self.updating = False

    def set_window_style(self):
        background_color = self.stage_model.get_node_attr_value(
            node_path=self.window_node_path,
            attr_name=self.WINDOW_COLOR_ATTR)
        background_color = background_color or '#232323'
        style = '''
                QFrame {
                    background-color:%s;
                    border-radius: 11px;
                    border: 0px solid transparent;
                }
                ''' % background_color
        self.window_frame.setStyleSheet(style)

    @staticmethod
    def set_tab_style(tab_widget):
        if not tab_widget:
            return

        if not tab_widget.count():
            return

        tab = tab_widget.currentWidget()
        tab.set_tab_style()

    def set_tab_index(self, tab_widget):
        data = self.widget_state_data.get(tab_widget.tab_widget_path)
        if not data:
            return

        index = tab_widget.currentIndex()
        self.widget_state_data[tab_widget.tab_widget_path]['state'] = index

    def tab_context_menu(self, tab_widget):
        cursor_pos = QtGui.QCursor.pos()
        tab_widget_pos = tab_widget.mapFromGlobal(cursor_pos)
        tab_index = tab_widget.tabBar().tabAt(tab_widget_pos)
        tab = tab_widget.widget(tab_index)
        if tab:
            tab.menu.show_menu(tab.mapFromGlobal(cursor_pos))
        else:
            self.window_frame.menu.show_menu(tab_widget_pos)


def get_resolved_attr_values(node_path, stage_model):
    attr_values = {}
    attr_names = stage_model.get_node_attr_names(node_path=node_path,
                                                 layer=stage_model.comp_layer)
    for attr_name in attr_names:
        attr_value = stage_model.get_node_attr_value(
            node_path=node_path,
            attr_name=attr_name,
            layer=stage_model.comp_layer,
            data_state=DATA_STATE.RESOLVED)
        attr_values[attr_name] = attr_value
    return attr_values


def get_widget_type(node_path, stage_model):
    node = stage_model.comp_layer.lookup(node_path)
    if not node:
        return

    instance_trace_list = stage_model.stage.get_instance_sources(
        node=node,
        trace_list=[],
        comp_layer=stage_model.comp_layer)
    if not instance_trace_list:
        return

    instance_node = instance_trace_list[-1]
    instance_path = stage_model.comp_layer.get_node_path(instance_node)
    if not instance_path:
        return

    instance_name = nxt_path.node_name_from_node_path(instance_path)
    widget_type = instance_name if instance_name in WIDGET_TYPES else None
    return widget_type


class ContextMenu(QtWidgets.QMenu):

    def __init__(self, stage_model, node_path, widget, items=None,
                 parent=None):
        super(ContextMenu, self).__init__(parent=parent)
        self.setStyleSheet(
            'background-color: #232323; border: 1px solid #3E3E3E')

        self.stage_model = stage_model
        self.node_path = node_path
        self.widget = widget

        # add context menu to widget
        self.widget.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.widget.customContextMenuRequested.connect(self.show_menu)
        self.widget.menu = self

        # add menu items
        items = items or []
        for item in items:
            if item == 'separator':
                self.addSeparator()
            else:
                self.addAction(item)

        # add node selection item
        if items:
            self.addSeparator()
        node_action = QtWidgets.QAction('Select NXT Node', self)
        node_action.triggered.connect(self.select_node)
        self.addAction(node_action)

    def show_menu(self, pos):
        self.exec_(self.widget.mapToGlobal(pos))

    def select_node(self):
        self.stage_model.select_and_frame(self.node_path)


class Tab(QtWidgets.QWidget):

    TEXT_ATTR = 'text'
    BACKGROUND_COLOR_ATTR = 'tab_color'
    BORDER_COLOR_ATTR = 'tab_border_color'

    def __init__(self, node_path, tab_parent, parent=None):
        super(Tab, self).__init__(parent=parent)

        # internal setup
        self._parent = parent
        self._tab_parent = tab_parent
        self.stage_model = parent.stage_model
        self.node_path = node_path

        attr_data = get_resolved_attr_values(node_path=node_path,
                                             stage_model=self.stage_model)

        self.background_color = attr_data.get(self.BACKGROUND_COLOR_ATTR)
        self.border_color = attr_data.get(self.BORDER_COLOR_ATTR)

        # setup tab
        text = attr_data.get(self.TEXT_ATTR) or self.TEXT_ATTR
        tab_parent.addTab(self, text)

        # main layout
        self.main_layout = QtWidgets.QVBoxLayout(self)
        self.main_layout.setContentsMargins(0, 0, 0, 0)
        self.main_layout.setSpacing(0)
        self.setLayout(self.main_layout)

        # scroll area
        self.scroll_area = QtWidgets.QScrollArea(self)
        self.scroll_area.setWidgetResizable(True)
        style = '''
                QScrollBar:vertical {
                    background-color: #303030;
                    width: 16px;
                    margin: 16px 2px 16px 2px;
                    border: none;
                    border-radius: 4px;
                }

                QScrollBar::handle:vertical {
                    background-color: #606060;
                    border: none;
                    min-height: 8px;
                    border-radius: 4px;
                }

                QScrollBar::handle:vertical:hover {
                    background-color: #148CD2;
                    border: none;
                    border-radius: 4px;
                    min-height: 8px;

                }
                '''
        self.scroll_area.verticalScrollBar().setStyleSheet(style)
        self.scroll_area.setStyleSheet(style)
        self.main_layout.addWidget(self.scroll_area)

        self.scroll_widget = QtWidgets.QWidget(self.scroll_area)
        self.scroll_area.setWidget(self.scroll_widget)

        # layout
        self.layout = QtWidgets.QVBoxLayout()
        self.layout.setContentsMargins(0, 0, 6, 0)
        self.layout.setSpacing(8)
        self.layout.setAlignment(QtCore.Qt.AlignTop)
        self.scroll_widget.setLayout(self.layout)

        # add context menu
        ContextMenu(stage_model=self.stage_model,
                    node_path=node_path,
                    widget=self,
                    items=None,
                    parent=self)

    def set_tab_style(self):
        background_color = self.background_color if self.background_color else '#3E3E3E'
        border = '2px solid ' + self.border_color if self.border_color else 'none'
        style = '''
                        QTabWidget::pane {
                            border-radius: 9px;
                            border: %s;
                            background-color: %s;
                        }

                        QTabWidget::tab-bar {
                            left: 5px;
                        }

                        QTabBar {
                            background-color: transparent;
                            border: none;
                        }

                        QTabBar::tab {
                            alignment: center;
                            border: %s;
                            border-radius: 8px;
                            padding-bottom: 14px;
                            margin-bottom: -10px;
                            background-color: %s;
                        }

                        QTabBar::tab:hover {
                            padding-top: 2px;
                        }

                        QTabBar::tab:!selected {
                            border: none;
                            background-color: #303030;
                        }

                        QTabBar::tab:!selected:hover {
                            padding-top: 0px;
                        }

                        QWidget {
                            background-color: transparent;
                        }
                        ''' % (border, background_color, border,
                               background_color)
        self._tab_parent.setStyleSheet(style)


class Panel(QtWidgets.QWidget):

    TEXT_ATTR = 'text'
    TEXT_COLOR_ATTR = 'text_color'
    BACKGROUND_COLOR_ATTR = 'panel_color'
    BORDER_COLOR_ATTR = 'panel_border_color'

    def __init__(self, node_path, parent=None):
        super(Panel, self).__init__(parent=parent)

        # internal setup
        self._parent = parent
        self.node_path = node_path
        self.stage_model = parent.stage_model

        attr_data = get_resolved_attr_values(node_path=node_path,
                                             stage_model=self.stage_model)

        # main layout
        self.main_layout = QtWidgets.QVBoxLayout(self)
        self.main_layout.setContentsMargins(0, 0, 0, 0)
        self.main_layout.setSpacing(0)
        self.main_layout.setAlignment(QtCore.Qt.AlignTop)
        self.setLayout(self.main_layout)

        # panel frame
        background_color = attr_data.get(self.BACKGROUND_COLOR_ATTR)
        border_color = attr_data.get(self.BORDER_COLOR_ATTR)
        background_color = background_color or '#3E3E3E'
        border_color = border_color or 'transparent'
        style = '''
                QFrame {
                    background-color: %s;
                    border-radius: 9px;
                    border: 2px solid %s;
                    }
                ''' % (background_color, border_color)
        self.panel_frame = QtWidgets.QFrame(self)
        self.panel_frame.setStyleSheet(style)
        self.main_layout.addWidget(self.panel_frame)

        # panel layout
        self.panel_layout = QtWidgets.QVBoxLayout(self.panel_frame)
        self.panel_layout.setContentsMargins(0, 0, 0, 0)
        self.panel_layout.setSpacing(0)
        self.panel_frame.setLayout(self.panel_layout)

        # title layout
        self.title_layout = QtWidgets.QHBoxLayout()
        self.title_layout.setContentsMargins(4, 0, 0, 0)
        self.title_layout.setSpacing(0)
        self.panel_layout.addLayout(self.title_layout)

        # collapse button
        self.collapse_button = QtWidgets.QToolButton(checkable=True,
                                                     checked=False)
        self.collapse_button.setStyleSheet(
            'QToolButton {background-color: transparent; border: none;}')
        self.collapse_button.setArrowType(QtCore.Qt.DownArrow)
        self.collapse_button.toggled.connect(self.toggle_collapsed)
        self.title_layout.addWidget(self.collapse_button)

        # title label
        text = attr_data.get(self.TEXT_ATTR) or self.TEXT_ATTR
        text_color = attr_data.get(self.TEXT_COLOR_ATTR)
        self.title_label = DoubleClickLabel(text=text, parent=self.panel_frame)
        if text_color:
            self.title_label.setStyleSheet(
                'QLabel {color: %s; border-color: none;}' % text_color)
        else:
            self.title_label.setStyleSheet(
                'QLabel {border-color: transparent;}')
        self.title_label.doubleClicked.connect(self.collapse_button.toggle)
        self.title_layout.addWidget(self.title_label)

        # main frame
        self.main_frame = QtWidgets.QFrame(self)
        self.main_frame.setStyleSheet('QFrame {border: none;}')
        self.panel_layout.addWidget(self.main_frame)

        # layout
        self.layout = QtWidgets.QVBoxLayout(self.main_frame)
        self.layout.setContentsMargins(10, 4, 10, 10)
        self.main_frame.setLayout(self.layout)

        ContextMenu(stage_model=self.stage_model,
                    node_path=node_path,
                    widget=self,
                    items=None,
                    parent=self)

        # set state
        data = parent.widget_state_data.get(node_path)
        if data:
            if data['widget_type'] == 'panel':
                state = data.get('state', False)
                self.set_collapse_state(state)
            else:
                data['widget_type'] = 'panel'
                data['state'] = self.is_collapsed
            data['widget'] = self
        else:
            data = {'widget_type': 'panel',
                    'widget': self,
                    'state': self.is_collapsed}
        parent.widget_state_data[self.node_path] = data

    @property
    def is_collapsed(self):
        return self.collapse_button.isChecked()

    def set_collapse_state(self, state):
        self.collapse_button.setChecked(state)

    def toggle_collapsed(self):
        if self.is_collapsed:
            self.collapse_button.setArrowType(QtCore.Qt.RightArrow)
            self.main_frame.hide()
        else:
            self.collapse_button.setArrowType(QtCore.Qt.DownArrow)
            self.main_frame.show()

        state = self.is_collapsed
        self._parent.widget_state_data[self.node_path]['state'] = state


class DoubleClickLabel(QtWidgets.QLabel):
    doubleClicked = QtCore.Signal()

    def __init__(self, *args, **kwargs):
        super(DoubleClickLabel, self).__init__(*args, **kwargs)

    def mouseDoubleClickEvent(self, event):
        if event.button() == QtCore.Qt.LeftButton:
            self.doubleClicked.emit()


class GridLayout(QtWidgets.QWidget):

    MAX_COLUMN_ATTR = 'max_columns'

    def __init__(self, node_path, parent=None):
        super(GridLayout, self).__init__(parent=parent)

        # internal setup
        self._parent = parent
        self.stage_model = parent.stage_model

        attr_data = get_resolved_attr_values(node_path=node_path,
                                             stage_model=self.stage_model)

        # layout
        self.layout = QtWidgets.QGridLayout()
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(self.layout)

        # set column stretch
        # this also also has the side effect of defining the number of columns
        # so no custom row/col logic is required when using "addWidget".
        max_columns = attr_data.get(self.MAX_COLUMN_ATTR)
        self.max_columns = int(max_columns) if max_columns else 1
        for c in range(self.max_columns):
            self.layout.setColumnStretch(c, 1)

        ContextMenu(stage_model=self.stage_model,
                    node_path=node_path,
                    widget=self,
                    items=None,
                    parent=self)


class Button(QtWidgets.QPushButton):

    MENU_ITEM_TYPE_ATTR = 'menuItem'
    TEXT_ATTR = 'text'
    TEXT_COLOR_ATTR = 'text_color'
    BACKGROUND_COLOR_ATTR = 'button_color'
    BORDER_COLOR_ATTR = 'button_border_color'
    EXEC_PATH_ATTR = 'button_exec_path'
    ITEM_PATH_ATTR = 'item_exec_path'
    SEPARATOR_ATTR = 'item_is_separator'
    SELECTOR_ITEM_ATTR = 'is_item_selector'
    ITEM_LIST_PATH_ATTR = 'item_list_attr_path'
    INPUT_LIST_ATTR = 'item_input_list'
    SELECTOR_TITLE_ATTR = 'item_selector_title'

    def __init__(self, node_path, parent=None):
        super(Button, self).__init__(parent=parent)

        # internal setup
        self._parent = parent
        self.stage_model = parent.stage_model
        self.node_path = node_path

        attr_data = get_resolved_attr_values(node_path=node_path,
                                             stage_model=self.stage_model)

        # button setup
        text = (attr_data.get(self.TEXT_ATTR) or
                nxt_path.node_name_from_node_path(node_path))
        self.setText(text)
        self.pressed.connect(lambda p=self.node_path, a=self.EXEC_PATH_ATTR:
                             self.execute_node_path(p, a))

        background_color = attr_data.get(self.BACKGROUND_COLOR_ATTR) or '#606060'
        border_color = attr_data.get(self.BORDER_COLOR_ATTR) or 'none'
        text_color = attr_data.get(self.TEXT_COLOR_ATTR) or 'white'
        style = '''
                QPushButton {
                    color: %s;
                    background-color: %s;
                    border-radius: 9px;
                    border: none;
                    }

                QPushButton:hover {
                    padding-bottom: -1;
                    padding-top: -1;
                    border: 2px solid %s;
                    }
                ''' % (text_color, background_color, border_color)
        self.setStyleSheet(style)

        # add context menu
        items = []
        for menu_item_path in self.get_menu_item_paths():
            menu_data = get_resolved_attr_values(node_path=menu_item_path,
                                                 stage_model=self.stage_model)
            if menu_data.get(self.SEPARATOR_ATTR) == 'True':
                items.append('separator')
            else:
                text = menu_data.get(self.TEXT_ATTR)
                action = QtWidgets.QAction(text, self)
                if menu_data.get(self.SELECTOR_ITEM_ATTR) == 'True':
                    title = menu_data.get(self.SELECTOR_TITLE_ATTR)
                    action.triggered.connect(lambda p=menu_item_path, t=title:
                                             self.selection_widget(p, t))
                else:
                    action.triggered.connect(lambda p=menu_item_path,
                                             a=self.ITEM_PATH_ATTR:
                                             self.execute_node_path(p, a))
                items.append(action)

        ContextMenu(stage_model=self.stage_model,
                    node_path=self.node_path,
                    widget=self,
                    items=items,
                    parent=self)

    def get_menu_item_paths(self):
        children = self.stage_model.get_children(
            node_path=self.node_path,
            layer=self.stage_model.comp_layer,
            ordered=True)

        menu_items = []
        for child in children:
            widget_type = get_widget_type(child, self.stage_model)
            if widget_type == self.MENU_ITEM_TYPE_ATTR:
                menu_items.append(child)

        return menu_items

    def selection_widget(self, node_path, title):
        # execute node to calculate input list
        if user_dir.user_prefs.get(RECOMP_PREF, True):
            rt_layer = None
        else:
            rt_layer = self.stage_model.current_rt_layer
        self.stage_model.execute_nodes(node_paths=[node_path],
                                       rt_layer=rt_layer)

        # get items list from input value
        # try resolved value first then fall back to cached value
        input_value = self.stage_model.get_node_attr_value(
            node_path=node_path,
            attr_name=self.INPUT_LIST_ATTR,
            layer=None,
            data_state=DATA_STATE.RESOLVED)
        if not input_value:
            input_value = self.stage_model.get_node_attr_value(
                node_path=node_path,
                attr_name=self.INPUT_LIST_ATTR,
                layer=None,
                data_state=DATA_STATE.CACHED)
        if (not isinstance(input_value, str) and
                isinstance(input_value, Iterable)):
            items = input_value
        else:
            items = []
        # selector dialog
        screen = QtWidgets.QApplication.desktop().screenNumber(
            QtWidgets.QApplication.desktop().cursor().pos())
        center = QtWidgets.QApplication.desktop().screenGeometry(screen).center()
        dialog = SelectionDialog(title=title, items=items, pos=center, parent=self)
        dialog.exec_()
        if not dialog.result():
            return

        list_path_attr = self.stage_model.get_node_attr_value(
            node_path=node_path,
            attr_name=self.ITEM_LIST_PATH_ATTR,
            layer=None)

        try:
            target_node, target_attr = list_path_attr.split('.')
        except ValueError:
            logger.error('You must enter a /valid/node/path.attr in "{}.{}",'
                         'Invalid: {}'
                         ''.format(node_path, self.ITEM_LIST_PATH_ATTR,
                                   list_path_attr))
            return
        self.stage_model.set_node_attr_value(
            node_path=target_node,
            attr_name=target_attr,
            value=dialog.item_list,
            layer=None)

    def execute_node_path(self, node_path, attr_name):
        exec_path = self.stage_model.get_node_attr_value(
            node_path=node_path,
            attr_name=attr_name,
            layer=self.stage_model.comp_layer,
            data_state=DATA_STATE.RESOLVED)

        if not exec_path:
            return

        descendants = self.stage_model.comp_layer.descendants(exec_path,
                                                              ordered=True)

        # filter out widget nodes
        filtered_nodes = []
        dont_run = []
        for path in descendants:
            widget_type = get_widget_type(path, self.stage_model)
            if widget_type:
                des = self.stage_model.comp_layer.descendants(path,
                                                              ordered=True)
                dont_run += des
            if not widget_type and path not in dont_run:
                enabled = self.stage_model.get_node_enabled(path)
                anc_enabled = self.stage_model.get_node_ancestor_enabled(path)
                if enabled and anc_enabled:
                    filtered_nodes.append(path)
                else:
                    des = self.stage_model.comp_layer.descendants(path,
                                                                  ordered=True)
                    dont_run += [path] + des

        node_paths = [exec_path] + filtered_nodes
        if user_dir.user_prefs.get(RECOMP_PREF, True):
            rt_layer = None
        else:
            rt_layer = self.stage_model.current_rt_layer
        self.stage_model.execute_nodes(node_paths=node_paths,
                                       rt_layer=rt_layer)


class CheckBox(QtWidgets.QWidget):

    TEXT_ATTR = 'text'
    TEXT_COLOR_ATTR = 'text_color'
    TEXT_ALIGN_ATTR = 'text_align'
    IS_CHECKED_ATTR = 'is_checked'
    ALIGNMENT_ATTR = 'checkbox_alignment'
    EXEC_PATH_ATTR = 'checkbox_exec_path'
    ALIGNMENT = {'center': QtCore.Qt.AlignHCenter,
                 'left': QtCore.Qt.AlignLeft,
                 'right': QtCore.Qt.AlignRight}
    LABEL_ALIGNMENT = {'left': QtCore.Qt.RightToLeft,
                       'right': QtCore.Qt.LeftToRight}

    def __init__(self, node_path, parent=None):
        super(CheckBox, self).__init__(parent=parent)
        self.setStyleSheet('QWidget {margin-top: -4; margin-bottom: -4;}')

        # internal setup
        self._parent = parent
        self.stage_model = parent.stage_model
        self.node_path = node_path

        attr_data = get_resolved_attr_values(node_path=node_path,
                                             stage_model=self.stage_model)

        # layout
        self.main_layout = QtWidgets.QHBoxLayout(self)
        self.main_layout.setContentsMargins(0, 0, 0, 0)
        self.main_layout.setSpacing(0)
        alignment = attr_data.get(self.ALIGNMENT_ATTR)
        layout_alignment = self.ALIGNMENT.get(alignment)
        if layout_alignment:
            self.main_layout.setAlignment(layout_alignment)
        self.setLayout(self.main_layout)

        # checkbox
        text_color = attr_data.get(self.TEXT_COLOR_ATTR) or 'white'
        is_checked = attr_data.get(self.IS_CHECKED_ATTR)
        checked = True if is_checked == 'True' else False
        text = (attr_data.get(self.TEXT_ATTR) or
                nxt_path.node_name_from_node_path(self.node_path))
        style = 'QCheckBox {color: %s;}' % text_color
        self.checkbox = QtWidgets.QCheckBox(text=text, checked=checked)
        self.checkbox.setStyleSheet(style)
        self.checkbox.toggled.connect(self.execute_node_path)
        self.main_layout.addWidget(self.checkbox)

        text_align = attr_data.get(self.TEXT_ALIGN_ATTR)
        label_alignment = self.LABEL_ALIGNMENT.get(text_align)
        if label_alignment:
            self.checkbox.setLayoutDirection(label_alignment)

        # add context menu
        ContextMenu(stage_model=self.stage_model,
                    node_path=node_path,
                    widget=self.checkbox,
                    items=None,
                    parent=self)

    def execute_node_path(self):
        check_state = 'True' if self.checkbox.checkState() else 'False'
        self.stage_model.set_node_attr_value(node_path=self.node_path,
                                             attr_name=self.IS_CHECKED_ATTR,
                                             value=check_state,
                                             layer=None)

        exec_path = self.stage_model.get_node_attr_value(
            node_path=self.node_path,
            attr_name=self.EXEC_PATH_ATTR,
            layer=self.stage_model.comp_layer,
            data_state=DATA_STATE.RESOLVED)
        if exec_path:
            if not self.stage_model.node_exists(node_path=self.node_path,
                                                layer=self.stage_model.comp_layer):
                return
            if user_dir.user_prefs.get(RECOMP_PREF, True):
                rt_layer = None
            else:
                rt_layer = self.stage_model.current_rt_layer
            self.stage_model.execute_nodes(node_paths=[exec_path],
                                           rt_layer=rt_layer)


class DropDownMenu(QtWidgets.QWidget):

    TEXT_ATTR = 'text'
    TEXT_COLOR_ATTR = 'text_color'
    BACKGROUND_COLOR_ATTR = 'menu_background_color'
    BORDER_COLOR_ATTR = 'menu_border_color'
    MENU_ITEMS_ATTR = 'menu_items'
    EXEC_PATH_ATTR = 'menu_exec_path'
    MENU_VALUE_ATTR = 'menu_value'

    def __init__(self, node_path, parent=None):
        super(DropDownMenu, self).__init__(parent=parent)
        size_policy = QtWidgets.QSizePolicy()
        size_policy.setHorizontalPolicy(QtWidgets.QSizePolicy.Preferred)
        size_policy.setHorizontalStretch(1)
        self.setSizePolicy(size_policy)

        # internal setup
        self._parent = parent
        self.stage_model = parent.stage_model
        self.node_path = node_path

        attr_data = get_resolved_attr_values(node_path=node_path,
                                             stage_model=self.stage_model)

        # layout
        self.main_layout = QtWidgets.QHBoxLayout(self)
        self.main_layout.setContentsMargins(0, 0, 0, 0)
        self.main_layout.setSpacing(0)
        self.setLayout(self.main_layout)

        # label
        text = (attr_data.get(self.TEXT_ATTR) or
                nxt_path.node_name_from_node_path(self.node_path))
        text_color = attr_data.get(self.TEXT_COLOR_ATTR) or 'white'
        self.label = QtWidgets.QLabel(text)
        self.label.setStyleSheet('QLabel {color: %s}' % text_color)
        self.main_layout.addWidget(self.label)

        # style
        background_color = attr_data.get(self.BACKGROUND_COLOR_ATTR) or '#232323'
        border_color = attr_data.get(self.BORDER_COLOR_ATTR) or 'transparent'
        border_line = '2px solid %s' % border_color if border_color != 'transparent' else 'none'
        padding = '4' if border_color != 'transparent' else '6'
        style = '''
                QComboBox {
                    color: %s;
                    background: %s;
                    border-radius: 9px;
                    border: none;
                    padding-left: 6;
                    overflow: hidden;
                    }

                QComboBox:hover {
                    padding-left: %s;
                    border: %s;
                    }
                ''' % (text_color, background_color, padding,
                       border_line)
        view_style = '''
                     QComboBox QAbstractItemView {
                         selection-color: white;
                         selection-background-color: #148CD2;
                         background: %s;
                         border-top-left-radius: 0px;
                         border-top-right-radius: 0px;
                         border-bottom-left-radius: 9px;
                         border-bottom-right-radius: 9px;
                         border-width: 2px;
                         border-style: solid;
                         border-color: transparent %s %s %s;
                         margin-right: 9px;
                         margin-left: 9px;
                         outline: 0;
                     }
                     ''' % (background_color, border_color, border_color,
                            border_color)

        # combo box
        self.combo_box = QtWidgets.QComboBox(self)
        self.combo_box.setStyleSheet(style)
        self.combo_box.view().window().setWindowFlags(
            QtCore.Qt.Popup |
            QtCore.Qt.FramelessWindowHint |
            QtCore.Qt.NoDropShadowWindowHint)
        self.combo_box.view().setStyleSheet(view_style)
        self.combo_box.view().window().setAttribute(QtCore.Qt.WA_TranslucentBackground)
        self.main_layout.addWidget(self.combo_box, 1)

        # menu items
        menu_items = attr_data.get(self.MENU_ITEMS_ATTR)
        if not menu_items:
            return

        menu_items_list = menu_items.split(' ')
        self.combo_box.addItems(menu_items_list)
        menu_value = self.stage_model.get_node_attr_value(
            node_path=self.node_path,
            attr_name=self.MENU_VALUE_ATTR,
            layer=self.stage_model.comp_layer,
            data_state=DATA_STATE.RESOLVED)
        if menu_value in menu_items_list:
            self.combo_box.setCurrentText(menu_value)
        self.combo_box.currentIndexChanged.connect(self.execute_node_path)

        # add context menus
        ContextMenu(stage_model=self.stage_model,
                    node_path=node_path,
                    widget=self,
                    items=None,
                    parent=self)

        ContextMenu(stage_model=self.stage_model,
                    node_path=node_path,
                    widget=self.combo_box,
                    items=None,
                    parent=self)

    def set_menu_value(self):
        self.stage_model.set_node_attr_value(node_path=self.node_path,
                                             attr_name=self.MENU_VALUE_ATTR,
                                             value=self.combo_box.currentText(),
                                             layer=None)

    def execute_node_path(self):
        self.set_menu_value()

        exec_path = self.stage_model.get_node_attr_value(
            node_path=self.node_path,
            attr_name=self.EXEC_PATH_ATTR,
            layer=self.stage_model.comp_layer,
            data_state=DATA_STATE.RESOLVED)

        if exec_path:
            if not self.stage_model.node_exists(node_path=self.node_path,
                                                layer=self.stage_model.comp_layer):
                return
            if user_dir.user_prefs.get(RECOMP_PREF, True):
                rt_layer = None
            else:
                rt_layer = self.stage_model.current_rt_layer
            self.stage_model.execute_nodes(node_paths=[exec_path],
                                           rt_layer=rt_layer)


class SelectionDialog(QtWidgets.QDialog):

    def __init__(self, title=None, items=None, pos=None, parent=None):
        super(SelectionDialog, self).__init__(parent=parent)

        self.setWindowFlags(
            self.windowFlags() & ~QtCore.Qt.WindowContextHelpButtonHint)
        self.setWindowTitle(title or 'Select Items')
        self.setStyleSheet('QDialog {background: #232323;}')

        self.main_layout = QtWidgets.QVBoxLayout()
        self.main_layout.setContentsMargins(8, 8, 8, 8)
        self.main_layout.setSpacing(0)
        self.setLayout(self.main_layout)

        # scroll area
        self.scroll_area = QtWidgets.QScrollArea(self)
        self.scroll_area.setWidgetResizable(True)
        self.main_layout.addWidget(self.scroll_area)

        self.scroll_widget = QtWidgets.QWidget(self)
        self.scroll_area.setWidget(self.scroll_widget)

        self.scroll_layout = QtWidgets.QVBoxLayout(self)
        self.scroll_layout.setContentsMargins(0, 0, 4, 0)
        self.scroll_layout.setSpacing(4)
        self.scroll_layout.setAlignment(QtCore.Qt.AlignTop)
        self.scroll_widget.setLayout(self.scroll_layout)

        # list view
        self.model = QtGui.QStandardItemModel()
        items = items if items is not None else []
        for item in items:
            i = QtGui.QStandardItem(str(item))
            i.item_value = item
            i.setFlags(QtCore.Qt.ItemIsUserCheckable | QtCore.Qt.ItemIsEnabled)
            i.setData(QtCore.Qt.Checked, QtCore.Qt.CheckStateRole)
            self.model.appendRow(i)

        self.view = QtWidgets.QListView()
        self.view.setModel(self.model)
        self.view.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.view.customContextMenuRequested.connect(self.context_menu)
        self.scroll_layout.addWidget(self.view)

        # buttons
        self.button_layout = QtWidgets.QHBoxLayout(self)
        self.button_layout.setContentsMargins(0, 8, 0, 4)
        self.button_layout.setSpacing(8)
        self.main_layout.addLayout(self.button_layout)

        style = 'QPushButton {border-radius: 9px; background-color: #606060;}'

        self.accept_button = QtWidgets.QPushButton('Accept')
        self.accept_button.setStyleSheet(style)
        self.accept_button.pressed.connect(self.accept)
        self.button_layout.addWidget(self.accept_button)

        self.cancel_button = QtWidgets.QPushButton('Cancel')
        self.cancel_button.setStyleSheet(style)
        self.cancel_button.pressed.connect(self.reject)
        self.button_layout.addWidget(self.cancel_button)

    @property
    def item_list(self):
        item_list = []
        for row_index in range(self.model.rowCount()):
            item = self.model.item(row_index, 0)
            if item.checkState():
                item_list.append(item.item_value)
        return item_list

    def select_all(self):
        for row_index in range(self.model.rowCount()):
            item = self.model.item(row_index, 0)
            item.setCheckState(QtCore.Qt.Checked)

    def select_none(self):
        for row_index in range(self.model.rowCount()):
            item = self.model.item(row_index, 0)
            item.setCheckState(QtCore.Qt.Unchecked)

    def context_menu(self):
        menu = QtWidgets.QMenu(self)
        menu.addAction('Select All', self.select_all)
        menu.addAction('Select None', self.select_none)
        menu.popup(QtGui.QCursor.pos())



""" Dialogs. They're too small to deserve their own files.
"""
# Built-in
import logging
import os

# External
from Qt import QtCore, QtWidgets, QtGui

# Internal
import nxt_editor
from nxt_editor import user_dir

logger = logging.getLogger(nxt_editor.LOGGER_NAME)


class FilePathPrefEditor(QtWidgets.QDialog):
    """A dialog that defaults to existing value for a given preference key,
    provides a interface to pick a new file, and holds a value until
    user confirms intention to set the preference, when that preference
    is saved with the user's chosen value.
    """
    def __init__(self, pref_key, file_filter="*"):
        super(FilePathPrefEditor, self).__init__()
        self.pref_key = pref_key
        self.file_filter = file_filter
        self.default_file = None  # populated in self.populate
        self.layout = QtWidgets.QVBoxLayout(self)
        self.line_layout = QtWidgets.QHBoxLayout()
        self.layout.addLayout(self.line_layout)

        pixmap = QtGui.QPixmap(':icons/icons/copy_resolved_12.png')
        icon = QtGui.QIcon(pixmap)
        self.file_button = QtWidgets.QPushButton(icon, '')
        self.file_button.released.connect(self.choose_file)
        self.line_layout.addWidget(self.file_button)
        self.line_edit = QtWidgets.QLineEdit('')
        self.line_layout.addWidget(self.line_edit)

        self.buttons_layout = QtWidgets.QHBoxLayout()
        self.layout.addLayout(self.buttons_layout)

        self.cancel_button = QtWidgets.QPushButton('Cancel')
        self.cancel_button.released.connect(self.close)
        self.buttons_layout.addWidget(self.cancel_button)

        self.save_button = QtWidgets.QPushButton('Save')
        self.save_button.setDefault(True)
        self.save_button.released.connect(self.save_pressed)
        self.buttons_layout.addWidget(self.save_button)

        self.populate()

    def populate(self):
        self.default_file = user_dir.user_prefs.get(self.pref_key, '')
        self.line_edit.setText(self.default_file)

    def save_pressed(self):
        user_dir.user_prefs[self.pref_key] = str(self.line_edit.text())
        self.done(True)

    def choose_file(self):
        current_text = self.line_edit.text()
        dialog_dir = os.getcwd()
        if current_text:
            current_dir = os.path.dirname(current_text)
            if current_dir:
                dialog_dir = current_dir
        elif self.default_file:
            current_dir = os.path.dirname(self.default_file)
            if current_dir:
                dialog_dir = current_dir
        file_filter = self.file_filter
        chosen_file = QtWidgets.QFileDialog.getOpenFileName(filter=file_filter,
                                                            dir=dialog_dir)[0]
        if chosen_file:
            self.line_edit.setText(chosen_file)

    def exec_(self):
        self.populate()
        return super(FilePathPrefEditor, self).exec_()


class NxtFileDialog(QtWidgets.QDialog):
    def __init__(self, file_filter=("*.nxt", "*.nxtb"), caption=''):
        super(NxtFileDialog, self).__init__()
        self.filter = file_filter
        self.filter_string = " ".join(self.filter)
        self.dir = ''
        self._file_path = ''
        self.mode = 'open'
        self.caption = caption
        self.setWindowTitle("{} nxt layer".format(self.mode))
        self.modes = {'open': self._open,
                      'save': self._save}
        self.vb_main_layout = QtWidgets.QVBoxLayout()
        self.setLayout(self.vb_main_layout)

        self.hb_file_browse = QtWidgets.QHBoxLayout()
        self.vb_main_layout.addLayout(self.hb_file_browse)
        # File path line edit
        self.l_file_path = QtWidgets.QLineEdit()
        self.l_file_path.setText(self.get_base_path(with_file_name=False))
        self.hb_file_browse.addWidget(self.l_file_path)
        # TODO: Get the completer to complete relative paths based on the
        #  self.dir path. Right now it starts completions from the file
        #  system root.
        self.fsm_file = QtWidgets.QFileSystemModel()
        self.fsm_file.setFilter(QtCore.QDir.AllDirs |
                                QtCore.QDir.NoDotAndDotDot |
                                QtCore.QDir.Files)
        self.fsm_file.setNameFilters(self.filter)
        self.fsm_file.setNameFilterDisables(False)
        self.fsm_file.setRootPath(self.dir)
        self.completer = QtWidgets.QCompleter()
        self.completer.setModel(self.fsm_file)
        self.idx = QtCore.QModelIndex(self.fsm_file.index(self.dir))
        self.l_file_path.setCompleter(self.completer)
        # Default file browser
        self.btn_browse = QtWidgets.QPushButton('Browse')
        self.btn_browse.clicked.connect(self.browse)
        self.hb_file_browse.addWidget(self.btn_browse)

        self.hb_options = QtWidgets.QHBoxLayout()
        self.vb_main_layout.addLayout(self.hb_options)
        self.pb_rel_path = QtWidgets.QPushButton('Convert to relative path')
        self.pb_rel_path.clicked.connect(self.convert_path)
        self.hb_options.addWidget(self.pb_rel_path)
        self.btn_accept = QtWidgets.QPushButton(self.mode.capitalize())
        self.btn_accept.setDefault(True)
        self.btn_accept.clicked.connect(self.accept)
        self.hb_options.addWidget(self.btn_accept)

    def get_file_path(self, base_dir, mode, **kwargs):
        if 'filter' in kwargs.keys():
            self.filter = kwargs['filter']
        if not base_dir:
            base_dir = os.getcwd()
        self.dir = base_dir
        self.mode = mode
        self.setWindowTitle("{} nxt layer".format(self.mode))
        if not os.path.isdir(self.dir):
            self.l_file_path.setText(self.dir)
        self.exec_()

    @classmethod
    def system_file_dialog(cls, base_dir=None, mode='open', **kwargs):
        fd = cls(caption=kwargs.get('caption', ''))
        if 'filter' in kwargs.keys():
            fd.filter = kwargs['filter']
        if not base_dir:
            base_dir = os.getcwd()
        fd.dir = base_dir
        fd.mode = mode
        fd.browse()
        return fd.file_path

    @classmethod
    def get_open_file_path(cls, base_dir=None, **kwargs):
        fd = cls()
        fd.get_file_path(base_dir, 'open', **kwargs)
        return fd.file_path

    @classmethod
    def get_save_file_path(cls, base_dir=None, **kwargs):
        fd = cls()
        fd.get_file_path(base_dir, 'save', **kwargs)
        return fd.file_path

    def browse(self):
        file_path = self.modes[self.mode]()
        self.file_path = file_path

    def get_base_path(self, with_file_name=True):
        line_text = self.l_file_path.text()
        base = None
        if line_text:
            base = os.path.basename(line_text)
        if not base and os.path.isdir(self.dir) and with_file_name:
            base = os.path.join(self.dir, 'untitled.nxt')
        else:
            base = self.dir
        return base

    def _open(self):
        # TODO: Make the options a user pref
        ff = self.filter_string
        base = self.get_base_path()
        caption = 'Open'
        if self.caption:
            caption = self.caption
        options = QtWidgets.QFileDialog.DontUseNativeDialog
        file_path = QtWidgets.QFileDialog.getOpenFileName(filter=ff,
                                                          dir=base,
                                                          caption=caption,
                                                          options=options)[0]
        if not os.path.isfile(file_path):
            return
        return file_path

    def _save(self):
        ff = self.filter_string
        base = self.get_base_path()
        caption = 'Save'
        if self.caption:
            caption = self.caption
        options = QtWidgets.QFileDialog.DontUseNativeDialog
        file_path = QtWidgets.QFileDialog.getSaveFileName(filter=ff,
                                                          dir=base,
                                                          caption=caption,
                                                          options=options)[0]
        if not os.path.isdir(os.path.dirname(file_path)):
            return
        return file_path

    @property
    def file_path(self):
        return self._file_path

    @file_path.setter
    def file_path(self, file_path=None):
        line_text = self.l_file_path.text()
        if not file_path and not line_text:
            self._file_path = None
            return
        if not file_path:
            file_path = line_text
        p, ext = os.path.splitext(file_path)
        if not ext:
            file_path += '.nxt'
        if line_text != file_path:
            self.l_file_path.setText(file_path)
        self._file_path = file_path

    def set_file_path(self):
        line_text = self.l_file_path.text()
        self.file_path = line_text

    def convert_path(self):
        input_path = self.l_file_path.text()
        common = os.path.commonprefix([self.dir, input_path])
        rel_path = os.path.relpath(input_path, common)
        result_file_path = rel_path
        self.file_path = result_file_path

    def accept(self):
        self.file_path = self.l_file_path.text()
        super(NxtFileDialog, self).accept()

    def reject(self):
        self._file_path = None
        super(NxtFileDialog, self).reject()


class NxtWarningDialog(QtWidgets.QDialog):
    def __init__(self, text, info, details=''):
        super(NxtWarningDialog, self).__init__()
        self.show_details = False
        self._detail_text = details
        self.build_widgets()
        self.setDetailedText(self._detail_text)
        self.setText(text)
        self.setInformativeText(info)
        self.setWindowFlags(QtCore.Qt.WindowStaysOnTopHint)

    def build_widgets(self):
        self.icon = QtWidgets.QLabel("Pretend this is\nan icon")
        pixmap = QtGui.QPixmap(':icons/icons/nxt_err_128.png')
        self.icon.setPixmap(pixmap)
        self.text_label = QtWidgets.QLabel()
        self.text_label.setTextFormat(QtCore.Qt.RichText)
        self.info_label = QtWidgets.QLabel()
        self.show_details_button = QtWidgets.QPushButton("Show Details")
        self.show_details_button.released.connect(self.on_details_toggle)
        self.confirm_button = QtWidgets.QPushButton("Ok")
        self.confirm_button.setDefault(True)
        self.confirm_button.released.connect(self.close)

        self.details_text = QtWidgets.QLabel("Details")
        int_flags = (
            QtCore.Qt.TextSelectableByMouse |
            QtCore.Qt.TextSelectableByKeyboard
        )
        self.details_text.setTextInteractionFlags(int_flags)
        self.details_text.setFrameShape(QtWidgets.QFrame.Panel)
        self.details_text.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.details_text.setLineWidth(1)

        self.details_line = QtWidgets.QFrame()
        self.details_line.setFrameShape(QtWidgets.QFrame.HLine)
        self.details_line.setFrameShadow(QtWidgets.QFrame.Sunken)

        self.copy_details_button = QtWidgets.QPushButton("Copy to clipboard")
        self.copy_details_button.released.connect(self.on_copy_details)

        self.save_details_button = QtWidgets.QPushButton("Save to file")
        self.save_details_button.released.connect(self.on_save_details)

        self.detail_buttons_layout = QtWidgets.QHBoxLayout()
        self.detail_buttons_layout.addStretch(streth=1)
        self.detail_buttons_layout.addWidget(self.save_details_button)
        self.detail_buttons_layout.addWidget(self.copy_details_button)

        self.details_layout = QtWidgets.QVBoxLayout()
        self.details_layout.addWidget(self.details_line)
        self.details_layout.addWidget(self.details_text)
        self.details_layout.addLayout(self.detail_buttons_layout)
        self.details_widget = QtWidgets.QWidget()
        self.details_widget.setLayout(self.details_layout)

        self.buttons_layout = QtWidgets.QHBoxLayout()
        self.buttons_layout.addWidget(self.show_details_button)
        self.buttons_layout.addWidget(self.confirm_button)
        self.top_right_layout = QtWidgets.QVBoxLayout()
        self.top_right_layout.addWidget(self.text_label)
        self.top_right_layout.addWidget(self.info_label)
        self.top_right_layout.addStretch(streth=1)
        self.top_right_layout.addLayout(self.buttons_layout)
        self.top_layout = QtWidgets.QHBoxLayout()
        self.top_layout.addWidget(self.icon)
        self.top_layout.addLayout(self.top_right_layout)
        self.main_layout = QtWidgets.QVBoxLayout()
        self.setLayout(self.main_layout)
        self.main_layout.addLayout(self.top_layout)
        self.main_layout.addWidget(self.details_widget)

        self._update_details()

    def setText(self, text):
        bold = '<b>' + text + '</b>'
        self.text_label.setText(bold)

    def setDetailedText(self, details):
        self._detail_text = details
        self._update_details()

    def setInformativeText(self, text):
        self.info_label.setText(text)

    def on_details_toggle(self):
        self.show_details = not self.show_details
        self._update_details()

    def on_copy_details(self):
        clipboard = QtWidgets.QApplication.clipboard()
        clipboard.setText(self._detail_text)

    def on_save_details(self):
        file_name = QtWidgets.QFileDialog.getSaveFileName()[0]
        _, ext = os.path.splitext(file_name)
        if not ext:
            file_name += '.txt'
        with open(file_name, 'w+') as fp:
            fp.write(self._detail_text)

    def _update_details(self):
        if self._detail_text:
            self.show_details_button.show()
            self.details_text.setText(self._detail_text)
        else:
            self.show_details_button.hide()
        if self.show_details:
            self.show_details_button.setText("Hide Details...")
            self.details_widget.show()
        else:
            self.show_details_button.setText("Show Details...")
            self.details_widget.hide()

    @classmethod
    def show_message(cls, text, info, details=None):
        dialog = cls(text=text, info=info, details=details)
        dialog.exec_()


class NxtConfirmDialog(QtWidgets.QMessageBox):
    def __init__(self, text='Title', info='Confirm something!',
                 button_text=None, icon=QtWidgets.QMessageBox.Icon.Question):
        """Simple message box used for user confirmation
        :param text: Title text
        :param info: Main info text
        :param button_text: Custom button text dict:
        {QtWidgets.QMessageBox.Ok: 'Custom Ok Text',
        QtWidgets.QMessageBox.Cancel: 'Custom Cancel Text'}
        """
        super(NxtConfirmDialog, self).__init__()
        self.setText(text)
        self.setInformativeText(info)
        self.setStandardButtons(self.Ok)
        self.setIcon(icon)
        self.setWindowFlags(QtCore.Qt.WindowStaysOnTopHint)
        self.setStandardButtons(self.Ok | self.Cancel)
        if button_text:
            self.setButtonText(self.Ok, button_text.get(self.Ok, 'Ok'))
            self.setButtonText(self.Cancel, button_text.get(self.Cancel,
                                                            'Cancel'))

    @classmethod
    def show_message(cls, *args, **kwargs):
        """Class method for easily showing the dialog. If the user clicks the
        'ok' button True is returned otherwise False is returned.
        :param kwargs: See __init__
        :return: bool
        """
        dialog = cls(*args, **kwargs)
        result = dialog.exec_()
        if result == dialog.Ok:
            return True
        return False


class UnsavedLayersDialogue(QtWidgets.QDialog):
    @classmethod
    def save_before_exit(cls, stage_models, main_window):
        inst = cls(stage_models, main_window)
        return inst.exec_()

    def __init__(self, stage_models, main_window):
        self.main_window = main_window
        super(UnsavedLayersDialogue, self).__init__(parent=main_window)
        self.setWindowTitle('Unsaved layers')
        self.setWhatsThis('Please check the box next the layers you would like '
                          'to save.')
        self.main_layout = QtWidgets.QVBoxLayout()
        self.setLayout(self.main_layout)
        # Message Label
        message = ("The checkable layers below have unsaved changes.\n"
                   "Select layers to save.")
        self.message_label = QtWidgets.QLabel(message)
        self.main_layout.addWidget(self.message_label)
        # Unsaved Tree
        self.unsaved_tree = QtWidgets.QTreeView()
        self.unsaved_tree.setHeaderHidden(True)
        self.unsaved_model = self.make_unsaved_model(stage_models)
        self.unsaved_tree.setModel(self.unsaved_model)
        self.unsaved_tree.expandAll()
        self.main_layout.addWidget(self.unsaved_tree)
        # Response Buttons
        self.response_layout = QtWidgets.QHBoxLayout()
        self.cancel_button = QtWidgets.QPushButton('Cancel')
        self.cancel_button.released.connect(self.reject)
        self.response_layout.addWidget(self.cancel_button)
        self.save_button = QtWidgets.QPushButton('Save Checked and Close')
        self.save_button.released.connect(self.on_save_released)
        self.response_layout.addWidget(self.save_button)
        self.ignore_button = QtWidgets.QPushButton('Ignore and Close')
        self.ignore_button.released.connect(self.accept)
        self.response_layout.addWidget(self.ignore_button)
        self.main_layout.addLayout(self.response_layout)

    @staticmethod
    def make_unsaved_model(stage_models):
        model = QtGui.QStandardItemModel()

        def r_add(stage_model, layer, parent_item, dirty):
            item = QtGui.QStandardItem(layer.get_alias())
            item.setData(layer)
            item.setEditable(False)
            if layer in dirty:
                item.setCheckable(True)
                item.setCheckState(QtCore.Qt.Checked)
            color_code = stage_model.get_layer_color(layer)
            item.setBackground(QtGui.QBrush(QtGui.QColor(color_code)))
            item.setForeground(QtGui.QBrush(QtCore.Qt.white))
            parent_item.appendRow(item)
            for lay_dict in layer.sub_layers:
                r_add(stage_model, lay_dict['layer'], item, dirty)
        for s_m in stage_models:
            r_add(s_m, s_m.top_layer, model, s_m.get_unsaved_changes())
        return model

    def on_save_released(self):
        layers_to_save = []

        def save_checked(item):
            unsaved_layer = item.data()
            if item.checkState() == QtCore.Qt.Checked:
                layers_to_save.insert(0, unsaved_layer)
            for i in range(item.rowCount()):
                save_checked(item.child(i))
        for i in range(self.unsaved_model.rowCount()):
            save_checked(self.unsaved_model.item(i))
        for unsaved_layer in layers_to_save:
            self.main_window.save_layer(unsaved_layer)
        self.done(QtWidgets.QDialog.Accepted)


class UnsavedChangesMessage(QtWidgets.QMessageBox):
    @classmethod
    def save_before_close(cls, header=None, info=None):
        header = header or 'Unsaved changes detected.'
        message = cls()
        message.setWindowTitle('Unsaved Changes!')
        message.setText(header)
        info = info or 'Do you want to save your changes?'
        message.setInformativeText(info)
        message.setStandardButtons(message.Save |
                                   message.Discard |
                                   message.Cancel)
        message.setIcon(message.Warning)
        return message.exec_()
